from math import *
from model_BKZ import *

log_infinity = 9999

STEPS_b = 5
STEPS_m = 5


class MLWEParameterSet:
    def __init__(self, n, d, m, k, q, distr="binomial"):
        self.n = n          # Ring Dimension
        self.d = d          # MLWE Dimension (over the ring)
        self.m = m          # Number of Ring-Samples
        self.k = k          # Error Parameter
        self.q = q          # Modulus
        self.distr = distr  # Type of distribution : binomial or uniform (parameter k)


def LWE_primal_cost(q, n, m, s, b, cost_svp=svp_classical, verbose=False):
    """ Return the cost of the primal attack using m samples and blocksize b (infinity = fail)
    """
    d = n + m
    delta = delta_BKZ(b)
    if verbose:
        print("Primal attacks uses block-size %d and %d samples"%(b, m))
    
    if s * sqrt(b) < BKZ_last_block_length(q, m, n, b):
        return cost_svp(b)
    else:
        return log_infinity


def LWE_dual_cost(q, n, m, s, b, cost_svp=svp_classical, verbose=False):
    """ Return the cost of the dual attack using m samples and blocksize b (infinity = fail)
    """
    d = n + m
    l = BKZ_first_length(q, n, m, b)
    tau = l * s / q
    log2_eps = - 2 * pi * pi * tau**2 / log(2)
    log2_R = max(0, - 2 * log2_eps - nvec_sieve(b))
    if verbose:
        print ("Dual attacks uses block-size %d and %d samples"%(b, m))
        print ("shortest vector used has length l=%.2f, q=%d, `l<q'= %d"%(l, q, l<q))
        print ("log2(epsilon) = %.2f, log2 nvector per run %.2f"%(log2_eps, nvec_sieve(b)))
    return cost_svp(b) + log2_R


def MLWE_optimize_attack(q, n, max_m, s, cost_attack=LWE_primal_cost, cost_svp=svp_classical, verbose=True):
    """ Find optimal parameters for a given attack
    """
    best_cost = log_infinity
    for b in range(50, n+max_m, STEPS_b):
        if cost_svp(b) > best_cost:
            break
        for m in range(max(5, b-n), max_m, STEPS_m):
            cost = cost_attack(q, n, m, s, b, cost_svp)
            if cost<best_cost:
                best_cost = cost
                best_m = m
                best_b = b

    cost_attack(q, n, best_m, s, best_b, cost_svp=svp_classical, verbose=verbose)
    return (best_m, best_b, best_cost)


def check_eq(m_pc, m_pq, m_pp):
    if (m_pc != m_pq):
        print("m and b not equals among the three models")
    if (m_pq != m_pp):
        print("m and b not equals among the three models")


def MLWE_summarize_attacks(ps):
    """ Create a report on the best primal and dual BKZ attacks on an MLWE instance
    """
    q = ps.q
    n = ps.n * ps.d
    max_m = ps.n * ps.m
    
    if ps.distr=="binomial":
        s = sqrt(ps.k /2.)
    elif ps.distr=="uniform":
        k = ps.k
        s = sqrt(sum([i**2 for i in range(-k, k+1)])/(2*k+1))
    else:
        raise ValueError("Unknown distribution "+ps.distr)
    
    (m_pc, b_pc, c_pc) = MLWE_optimize_attack(q, n, max_m, s, cost_attack=LWE_primal_cost, cost_svp=svp_classical, verbose=True)
    (m_pq, b_pq, c_pq) = MLWE_optimize_attack(q, n, max_m, s, cost_attack=LWE_primal_cost, cost_svp=svp_quantum, verbose=False)
    (m_pp, b_pp, c_pp) = MLWE_optimize_attack(q, n, max_m, s, cost_attack=LWE_primal_cost, cost_svp=svp_plausible, verbose=False)

    check_eq(m_pc, m_pq, m_pp)
    check_eq(b_pc, b_pq, b_pp)

    print("Primal & %d & %d & %d & %d & %d"%(m_pq, b_pq, int(floor(c_pc)), int(floor(c_pq)), int(floor(c_pp))))

    (m_pc, b_pc, c_pc) = MLWE_optimize_attack(q, n, max_m, s, cost_attack=LWE_dual_cost, cost_svp=svp_classical, verbose=True)
    (m_pq, b_pq, c_pq) = MLWE_optimize_attack(q, n, max_m, s, cost_attack=LWE_dual_cost, cost_svp=svp_quantum, verbose=False)
    (m_pp, b_pp, c_pp) = MLWE_optimize_attack(q, n, max_m, s, cost_attack=LWE_dual_cost, cost_svp=svp_plausible, verbose=False)

    check_eq(m_pc, m_pq, m_pp)
    check_eq(b_pc, b_pq, b_pp)

    print("Dual & %d & %d & %d & %d & %d "%(m_pq, b_pq, int(floor(c_pc)), int(floor(c_pq)), int(floor(c_pp))))
    return (b_pq, int(floor(c_pc)), int(floor(c_pq)), int(floor(c_pp)))



# This file was automatically generated by SWIG (http://www.swig.org).
# Version 3.0.8
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.





from sys import version_info
if version_info >= (2, 6, 0):
    def swig_import_helper():
        from os.path import dirname
        import imp
        fp = None
        try:
            fp, pathname, description = imp.find_module('_kyber', [dirname(__file__)])
        except ImportError:
            import _kyber
            return _kyber
        if fp is not None:
            try:
                _mod = imp.load_module('_kyber', fp, pathname, description)
            finally:
                fp.close()
            return _mod
    _kyber = swig_import_helper()
    del swig_import_helper
else:
    import _kyber
del version_info
try:
    _swig_property = property
except NameError:
    pass  # Python < 2.2 doesn't have 'property'.


def _swig_setattr_nondynamic(self, class_type, name, value, static=1):
    if (name == "thisown"):
        return self.this.own(value)
    if (name == "this"):
        if type(value).__name__ == 'SwigPyObject':
            self.__dict__[name] = value
            return
    method = class_type.__swig_setmethods__.get(name, None)
    if method:
        return method(self, value)
    if (not static):
        if _newclass:
            object.__setattr__(self, name, value)
        else:
            self.__dict__[name] = value
    else:
        raise AttributeError("You cannot add attributes to %s" % self)


def _swig_setattr(self, class_type, name, value):
    return _swig_setattr_nondynamic(self, class_type, name, value, 0)


def _swig_getattr_nondynamic(self, class_type, name, static=1):
    if (name == "thisown"):
        return self.this.own()
    method = class_type.__swig_getmethods__.get(name, None)
    if method:
        return method(self)
    if (not static):
        return object.__getattr__(self, name)
    else:
        raise AttributeError(name)

def _swig_getattr(self, class_type, name):
    return _swig_getattr_nondynamic(self, class_type, name, 0)


def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)

try:
    _object = object
    _newclass = 1
except AttributeError:
    class _object:
        pass
    _newclass = 0



def cdata(ptr, nelements=1):
    return _kyber.cdata(ptr, nelements)
cdata = _kyber.cdata

def memmove(data, indata):
    return _kyber.memmove(data, indata)
memmove = _kyber.memmove
class SwigPyIterator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, SwigPyIterator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, SwigPyIterator, name)

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined - class is abstract")
    __repr__ = _swig_repr
    __swig_destroy__ = _kyber.delete_SwigPyIterator
    __del__ = lambda self: None

    def value(self):
        return _kyber.SwigPyIterator_value(self)

    def incr(self, n=1):
        return _kyber.SwigPyIterator_incr(self, n)

    def decr(self, n=1):
        return _kyber.SwigPyIterator_decr(self, n)

    def distance(self, x):
        return _kyber.SwigPyIterator_distance(self, x)

    def equal(self, x):
        return _kyber.SwigPyIterator_equal(self, x)

    def copy(self):
        return _kyber.SwigPyIterator_copy(self)

    def next(self):
        return _kyber.SwigPyIterator_next(self)

    def __next__(self):
        return _kyber.SwigPyIterator___next__(self)

    def previous(self):
        return _kyber.SwigPyIterator_previous(self)

    def advance(self, n):
        return _kyber.SwigPyIterator_advance(self, n)

    def __eq__(self, x):
        return _kyber.SwigPyIterator___eq__(self, x)

    def __ne__(self, x):
        return _kyber.SwigPyIterator___ne__(self, x)

    def __iadd__(self, n):
        return _kyber.SwigPyIterator___iadd__(self, n)

    def __isub__(self, n):
        return _kyber.SwigPyIterator___isub__(self, n)

    def __add__(self, n):
        return _kyber.SwigPyIterator___add__(self, n)

    def __sub__(self, *args):
        return _kyber.SwigPyIterator___sub__(self, *args)
    def __iter__(self):
        return self
SwigPyIterator_swigregister = _kyber.SwigPyIterator_swigregister
SwigPyIterator_swigregister(SwigPyIterator)


_kyber.SHARED_PTR_DISOWN_swigconstant(_kyber)
SHARED_PTR_DISOWN = _kyber.SHARED_PTR_DISOWN
class ucharCArray(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, ucharCArray, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, ucharCArray, name)
    __repr__ = _swig_repr

    def __init__(self, nelements):
        this = _kyber.new_ucharCArray(nelements)
        try:
            self.this.append(this)
        except Exception:
            self.this = this
    __swig_destroy__ = _kyber.delete_ucharCArray
    __del__ = lambda self: None

    def __getitem__(self, index):
        return _kyber.ucharCArray___getitem__(self, index)

    def __setitem__(self, index, value):
        return _kyber.ucharCArray___setitem__(self, index, value)

    def cast(self):
        return _kyber.ucharCArray_cast(self)
    __swig_getmethods__["frompointer"] = lambda x: _kyber.ucharCArray_frompointer
    if _newclass:
        frompointer = staticmethod(_kyber.ucharCArray_frompointer)
ucharCArray_swigregister = _kyber.ucharCArray_swigregister
ucharCArray_swigregister(ucharCArray)

def ucharCArray_frompointer(t):
    return _kyber.ucharCArray_frompointer(t)
ucharCArray_frompointer = _kyber.ucharCArray_frompointer

class uintCArray(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, uintCArray, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, uintCArray, name)
    __repr__ = _swig_repr

    def __init__(self, nelements):
        this = _kyber.new_uintCArray(nelements)
        try:
            self.this.append(this)
        except Exception:
            self.this = this
    __swig_destroy__ = _kyber.delete_uintCArray
    __del__ = lambda self: None

    def __getitem__(self, index):
        return _kyber.uintCArray___getitem__(self, index)

    def __setitem__(self, index, value):
        return _kyber.uintCArray___setitem__(self, index, value)

    def cast(self):
        return _kyber.uintCArray_cast(self)
    __swig_getmethods__["frompointer"] = lambda x: _kyber.uintCArray_frompointer
    if _newclass:
        frompointer = staticmethod(_kyber.uintCArray_frompointer)
uintCArray_swigregister = _kyber.uintCArray_swigregister
uintCArray_swigregister(uintCArray)

def uintCArray_frompointer(t):
    return _kyber.uintCArray_frompointer(t)
uintCArray_frompointer = _kyber.uintCArray_frompointer

class uint32CArray(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, uint32CArray, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, uint32CArray, name)
    __repr__ = _swig_repr

    def __init__(self, nelements):
        this = _kyber.new_uint32CArray(nelements)
        try:
            self.this.append(this)
        except Exception:
            self.this = this
    __swig_destroy__ = _kyber.delete_uint32CArray
    __del__ = lambda self: None

    def __getitem__(self, index):
        return _kyber.uint32CArray___getitem__(self, index)

    def __setitem__(self, index, value):
        return _kyber.uint32CArray___setitem__(self, index, value)

    def cast(self):
        return _kyber.uint32CArray_cast(self)
    __swig_getmethods__["frompointer"] = lambda x: _kyber.uint32CArray_frompointer
    if _newclass:
        frompointer = staticmethod(_kyber.uint32CArray_frompointer)
uint32CArray_swigregister = _kyber.uint32CArray_swigregister
uint32CArray_swigregister(uint32CArray)

def uint32CArray_frompointer(t):
    return _kyber.uint32CArray_frompointer(t)
uint32CArray_frompointer = _kyber.uint32CArray_frompointer

class intVector(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, intVector, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, intVector, name)
    __repr__ = _swig_repr

    def iterator(self):
        return _kyber.intVector_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self):
        return _kyber.intVector___nonzero__(self)

    def __bool__(self):
        return _kyber.intVector___bool__(self)

    def __len__(self):
        return _kyber.intVector___len__(self)

    def __getslice__(self, i, j):
        return _kyber.intVector___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _kyber.intVector___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _kyber.intVector___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _kyber.intVector___delitem__(self, *args)

    def __getitem__(self, *args):
        return _kyber.intVector___getitem__(self, *args)

    def __setitem__(self, *args):
        return _kyber.intVector___setitem__(self, *args)

    def pop(self):
        return _kyber.intVector_pop(self)

    def append(self, x):
        return _kyber.intVector_append(self, x)

    def empty(self):
        return _kyber.intVector_empty(self)

    def size(self):
        return _kyber.intVector_size(self)

    def swap(self, v):
        return _kyber.intVector_swap(self, v)

    def begin(self):
        return _kyber.intVector_begin(self)

    def end(self):
        return _kyber.intVector_end(self)

    def rbegin(self):
        return _kyber.intVector_rbegin(self)

    def rend(self):
        return _kyber.intVector_rend(self)

    def clear(self):
        return _kyber.intVector_clear(self)

    def get_allocator(self):
        return _kyber.intVector_get_allocator(self)

    def pop_back(self):
        return _kyber.intVector_pop_back(self)

    def erase(self, *args):
        return _kyber.intVector_erase(self, *args)

    def __init__(self, *args):
        this = _kyber.new_intVector(*args)
        try:
            self.this.append(this)
        except Exception:
            self.this = this

    def push_back(self, x):
        return _kyber.intVector_push_back(self, x)

    def front(self):
        return _kyber.intVector_front(self)

    def back(self):
        return _kyber.intVector_back(self)

    def assign(self, n, x):
        return _kyber.intVector_assign(self, n, x)

    def resize(self, *args):
        return _kyber.intVector_resize(self, *args)

    def insert(self, *args):
        return _kyber.intVector_insert(self, *args)

    def reserve(self, n):
        return _kyber.intVector_reserve(self, n)

    def capacity(self):
        return _kyber.intVector_capacity(self)
    __swig_destroy__ = _kyber.delete_intVector
    __del__ = lambda self: None
intVector_swigregister = _kyber.intVector_swigregister
intVector_swigregister(intVector)

class uintVector(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, uintVector, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, uintVector, name)
    __repr__ = _swig_repr

    def iterator(self):
        return _kyber.uintVector_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self):
        return _kyber.uintVector___nonzero__(self)

    def __bool__(self):
        return _kyber.uintVector___bool__(self)

    def __len__(self):
        return _kyber.uintVector___len__(self)

    def __getslice__(self, i, j):
        return _kyber.uintVector___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _kyber.uintVector___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _kyber.uintVector___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _kyber.uintVector___delitem__(self, *args)

    def __getitem__(self, *args):
        return _kyber.uintVector___getitem__(self, *args)

    def __setitem__(self, *args):
        return _kyber.uintVector___setitem__(self, *args)

    def pop(self):
        return _kyber.uintVector_pop(self)

    def append(self, x):
        return _kyber.uintVector_append(self, x)

    def empty(self):
        return _kyber.uintVector_empty(self)

    def size(self):
        return _kyber.uintVector_size(self)

    def swap(self, v):
        return _kyber.uintVector_swap(self, v)

    def begin(self):
        return _kyber.uintVector_begin(self)

    def end(self):
        return _kyber.uintVector_end(self)

    def rbegin(self):
        return _kyber.uintVector_rbegin(self)

    def rend(self):
        return _kyber.uintVector_rend(self)

    def clear(self):
        return _kyber.uintVector_clear(self)

    def get_allocator(self):
        return _kyber.uintVector_get_allocator(self)

    def pop_back(self):
        return _kyber.uintVector_pop_back(self)

    def erase(self, *args):
        return _kyber.uintVector_erase(self, *args)

    def __init__(self, *args):
        this = _kyber.new_uintVector(*args)
        try:
            self.this.append(this)
        except Exception:
            self.this = this

    def push_back(self, x):
        return _kyber.uintVector_push_back(self, x)

    def front(self):
        return _kyber.uintVector_front(self)

    def back(self):
        return _kyber.uintVector_back(self)

    def assign(self, n, x):
        return _kyber.uintVector_assign(self, n, x)

    def resize(self, *args):
        return _kyber.uintVector_resize(self, *args)

    def insert(self, *args):
        return _kyber.uintVector_insert(self, *args)

    def reserve(self, n):
        return _kyber.uintVector_reserve(self, n)

    def capacity(self):
        return _kyber.uintVector_capacity(self)
    __swig_destroy__ = _kyber.delete_uintVector
    __del__ = lambda self: None
uintVector_swigregister = _kyber.uintVector_swigregister
uintVector_swigregister(uintVector)

class ucharVector(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, ucharVector, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, ucharVector, name)
    __repr__ = _swig_repr

    def iterator(self):
        return _kyber.ucharVector_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self):
        return _kyber.ucharVector___nonzero__(self)

    def __bool__(self):
        return _kyber.ucharVector___bool__(self)

    def __len__(self):
        return _kyber.ucharVector___len__(self)

    def __getslice__(self, i, j):
        return _kyber.ucharVector___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _kyber.ucharVector___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _kyber.ucharVector___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _kyber.ucharVector___delitem__(self, *args)

    def __getitem__(self, *args):
        return _kyber.ucharVector___getitem__(self, *args)

    def __setitem__(self, *args):
        return _kyber.ucharVector___setitem__(self, *args)

    def pop(self):
        return _kyber.ucharVector_pop(self)

    def append(self, x):
        return _kyber.ucharVector_append(self, x)

    def empty(self):
        return _kyber.ucharVector_empty(self)

    def size(self):
        return _kyber.ucharVector_size(self)

    def swap(self, v):
        return _kyber.ucharVector_swap(self, v)

    def begin(self):
        return _kyber.ucharVector_begin(self)

    def end(self):
        return _kyber.ucharVector_end(self)

    def rbegin(self):
        return _kyber.ucharVector_rbegin(self)

    def rend(self):
        return _kyber.ucharVector_rend(self)

    def clear(self):
        return _kyber.ucharVector_clear(self)

    def get_allocator(self):
        return _kyber.ucharVector_get_allocator(self)

    def pop_back(self):
        return _kyber.ucharVector_pop_back(self)

    def erase(self, *args):
        return _kyber.ucharVector_erase(self, *args)

    def __init__(self, *args):
        this = _kyber.new_ucharVector(*args)
        try:
            self.this.append(this)
        except Exception:
            self.this = this

    def push_back(self, x):
        return _kyber.ucharVector_push_back(self, x)

    def front(self):
        return _kyber.ucharVector_front(self)

    def back(self):
        return _kyber.ucharVector_back(self)

    def assign(self, n, x):
        return _kyber.ucharVector_assign(self, n, x)

    def resize(self, *args):
        return _kyber.ucharVector_resize(self, *args)

    def insert(self, *args):
        return _kyber.ucharVector_insert(self, *args)

    def reserve(self, n):
        return _kyber.ucharVector_reserve(self, n)

    def capacity(self):
        return _kyber.ucharVector_capacity(self)
    __swig_destroy__ = _kyber.delete_ucharVector
    __del__ = lambda self: None
ucharVector_swigregister = _kyber.ucharVector_swigregister
ucharVector_swigregister(ucharVector)

class charVector(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, charVector, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, charVector, name)
    __repr__ = _swig_repr

    def iterator(self):
        return _kyber.charVector_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self):
        return _kyber.charVector___nonzero__(self)

    def __bool__(self):
        return _kyber.charVector___bool__(self)

    def __len__(self):
        return _kyber.charVector___len__(self)

    def __getslice__(self, i, j):
        return _kyber.charVector___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _kyber.charVector___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _kyber.charVector___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _kyber.charVector___delitem__(self, *args)

    def __getitem__(self, *args):
        return _kyber.charVector___getitem__(self, *args)

    def __setitem__(self, *args):
        return _kyber.charVector___setitem__(self, *args)

    def pop(self):
        return _kyber.charVector_pop(self)

    def append(self, x):
        return _kyber.charVector_append(self, x)

    def empty(self):
        return _kyber.charVector_empty(self)

    def size(self):
        return _kyber.charVector_size(self)

    def swap(self, v):
        return _kyber.charVector_swap(self, v)

    def begin(self):
        return _kyber.charVector_begin(self)

    def end(self):
        return _kyber.charVector_end(self)

    def rbegin(self):
        return _kyber.charVector_rbegin(self)

    def rend(self):
        return _kyber.charVector_rend(self)

    def clear(self):
        return _kyber.charVector_clear(self)

    def get_allocator(self):
        return _kyber.charVector_get_allocator(self)

    def pop_back(self):
        return _kyber.charVector_pop_back(self)

    def erase(self, *args):
        return _kyber.charVector_erase(self, *args)

    def __init__(self, *args):
        this = _kyber.new_charVector(*args)
        try:
            self.this.append(this)
        except Exception:
            self.this = this

    def push_back(self, x):
        return _kyber.charVector_push_back(self, x)

    def front(self):
        return _kyber.charVector_front(self)

    def back(self):
        return _kyber.charVector_back(self)

    def assign(self, n, x):
        return _kyber.charVector_assign(self, n, x)

    def resize(self, *args):
        return _kyber.charVector_resize(self, *args)

    def insert(self, *args):
        return _kyber.charVector_insert(self, *args)

    def reserve(self, n):
        return _kyber.charVector_reserve(self, n)

    def capacity(self):
        return _kyber.charVector_capacity(self)
    __swig_destroy__ = _kyber.delete_charVector
    __del__ = lambda self: None
charVector_swigregister = _kyber.charVector_swigregister
charVector_swigregister(charVector)

class doubleVector(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, doubleVector, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, doubleVector, name)
    __repr__ = _swig_repr

    def iterator(self):
        return _kyber.doubleVector_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self):
        return _kyber.doubleVector___nonzero__(self)

    def __bool__(self):
        return _kyber.doubleVector___bool__(self)

    def __len__(self):
        return _kyber.doubleVector___len__(self)

    def __getslice__(self, i, j):
        return _kyber.doubleVector___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _kyber.doubleVector___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _kyber.doubleVector___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _kyber.doubleVector___delitem__(self, *args)

    def __getitem__(self, *args):
        return _kyber.doubleVector___getitem__(self, *args)

    def __setitem__(self, *args):
        return _kyber.doubleVector___setitem__(self, *args)

    def pop(self):
        return _kyber.doubleVector_pop(self)

    def append(self, x):
        return _kyber.doubleVector_append(self, x)

    def empty(self):
        return _kyber.doubleVector_empty(self)

    def size(self):
        return _kyber.doubleVector_size(self)

    def swap(self, v):
        return _kyber.doubleVector_swap(self, v)

    def begin(self):
        return _kyber.doubleVector_begin(self)

    def end(self):
        return _kyber.doubleVector_end(self)

    def rbegin(self):
        return _kyber.doubleVector_rbegin(self)

    def rend(self):
        return _kyber.doubleVector_rend(self)

    def clear(self):
        return _kyber.doubleVector_clear(self)

    def get_allocator(self):
        return _kyber.doubleVector_get_allocator(self)

    def pop_back(self):
        return _kyber.doubleVector_pop_back(self)

    def erase(self, *args):
        return _kyber.doubleVector_erase(self, *args)

    def __init__(self, *args):
        this = _kyber.new_doubleVector(*args)
        try:
            self.this.append(this)
        except Exception:
            self.this = this

    def push_back(self, x):
        return _kyber.doubleVector_push_back(self, x)

    def front(self):
        return _kyber.doubleVector_front(self)

    def back(self):
        return _kyber.doubleVector_back(self)

    def assign(self, n, x):
        return _kyber.doubleVector_assign(self, n, x)

    def resize(self, *args):
        return _kyber.doubleVector_resize(self, *args)

    def insert(self, *args):
        return _kyber.doubleVector_insert(self, *args)

    def reserve(self, n):
        return _kyber.doubleVector_reserve(self, n)

    def capacity(self):
        return _kyber.doubleVector_capacity(self)
    __swig_destroy__ = _kyber.delete_doubleVector
    __del__ = lambda self: None
doubleVector_swigregister = _kyber.doubleVector_swigregister
doubleVector_swigregister(doubleVector)

class _string_list(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, _string_list, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, _string_list, name)
    __repr__ = _swig_repr

    def iterator(self):
        return _kyber._string_list_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self):
        return _kyber._string_list___nonzero__(self)

    def __bool__(self):
        return _kyber._string_list___bool__(self)

    def __len__(self):
        return _kyber._string_list___len__(self)

    def __getslice__(self, i, j):
        return _kyber._string_list___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _kyber._string_list___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _kyber._string_list___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _kyber._string_list___delitem__(self, *args)

    def __getitem__(self, *args):
        return _kyber._string_list___getitem__(self, *args)

    def __setitem__(self, *args):
        return _kyber._string_list___setitem__(self, *args)

    def pop(self):
        return _kyber._string_list_pop(self)

    def append(self, x):
        return _kyber._string_list_append(self, x)

    def empty(self):
        return _kyber._string_list_empty(self)

    def size(self):
        return _kyber._string_list_size(self)

    def swap(self, v):
        return _kyber._string_list_swap(self, v)

    def begin(self):
        return _kyber._string_list_begin(self)

    def end(self):
        return _kyber._string_list_end(self)

    def rbegin(self):
        return _kyber._string_list_rbegin(self)

    def rend(self):
        return _kyber._string_list_rend(self)

    def clear(self):
        return _kyber._string_list_clear(self)

    def get_allocator(self):
        return _kyber._string_list_get_allocator(self)

    def pop_back(self):
        return _kyber._string_list_pop_back(self)

    def erase(self, *args):
        return _kyber._string_list_erase(self, *args)

    def __init__(self, *args):
        this = _kyber.new__string_list(*args)
        try:
            self.this.append(this)
        except Exception:
            self.this = this

    def push_back(self, x):
        return _kyber._string_list_push_back(self, x)

    def front(self):
        return _kyber._string_list_front(self)

    def back(self):
        return _kyber._string_list_back(self)

    def assign(self, n, x):
        return _kyber._string_list_assign(self, n, x)

    def resize(self, *args):
        return _kyber._string_list_resize(self, *args)

    def insert(self, *args):
        return _kyber._string_list_insert(self, *args)

    def reserve(self, n):
        return _kyber._string_list_reserve(self, n)

    def capacity(self):
        return _kyber._string_list_capacity(self)
    __swig_destroy__ = _kyber.delete__string_list
    __del__ = lambda self: None
_string_list_swigregister = _kyber._string_list_swigregister
_string_list_swigregister(_string_list)

class _string_list_list(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, _string_list_list, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, _string_list_list, name)
    __repr__ = _swig_repr

    def iterator(self):
        return _kyber._string_list_list_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self):
        return _kyber._string_list_list___nonzero__(self)

    def __bool__(self):
        return _kyber._string_list_list___bool__(self)

    def __len__(self):
        return _kyber._string_list_list___len__(self)

    def __getslice__(self, i, j):
        return _kyber._string_list_list___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _kyber._string_list_list___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _kyber._string_list_list___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _kyber._string_list_list___delitem__(self, *args)

    def __getitem__(self, *args):
        return _kyber._string_list_list___getitem__(self, *args)

    def __setitem__(self, *args):
        return _kyber._string_list_list___setitem__(self, *args)

    def pop(self):
        return _kyber._string_list_list_pop(self)

    def append(self, x):
        return _kyber._string_list_list_append(self, x)

    def empty(self):
        return _kyber._string_list_list_empty(self)

    def size(self):
        return _kyber._string_list_list_size(self)

    def swap(self, v):
        return _kyber._string_list_list_swap(self, v)

    def begin(self):
        return _kyber._string_list_list_begin(self)

    def end(self):
        return _kyber._string_list_list_end(self)

    def rbegin(self):
        return _kyber._string_list_list_rbegin(self)

    def rend(self):
        return _kyber._string_list_list_rend(self)

    def clear(self):
        return _kyber._string_list_list_clear(self)

    def get_allocator(self):
        return _kyber._string_list_list_get_allocator(self)

    def pop_back(self):
        return _kyber._string_list_list_pop_back(self)

    def erase(self, *args):
        return _kyber._string_list_list_erase(self, *args)

    def __init__(self, *args):
        this = _kyber.new__string_list_list(*args)
        try:
            self.this.append(this)
        except Exception:
            self.this = this

    def push_back(self, x):
        return _kyber._string_list_list_push_back(self, x)

    def front(self):
        return _kyber._string_list_list_front(self)

    def back(self):
        return _kyber._string_list_list_back(self)

    def assign(self, n, x):
        return _kyber._string_list_list_assign(self, n, x)

    def resize(self, *args):
        return _kyber._string_list_list_resize(self, *args)

    def insert(self, *args):
        return _kyber._string_list_list_insert(self, *args)

    def reserve(self, n):
        return _kyber._string_list_list_reserve(self, n)

    def capacity(self):
        return _kyber._string_list_list_capacity(self)
    __swig_destroy__ = _kyber.delete__string_list_list
    __del__ = lambda self: None
_string_list_list_swigregister = _kyber._string_list_list_swigregister
_string_list_list_swigregister(_string_list_list)


_kyber.KYBER_K_swigconstant(_kyber)
KYBER_K = _kyber.KYBER_K

_kyber.CRYPTO_ALGNAME_swigconstant(_kyber)
CRYPTO_ALGNAME = _kyber.CRYPTO_ALGNAME

def crypto_kem_keypair(pk, sk):
    return _kyber.crypto_kem_keypair(pk, sk)
crypto_kem_keypair = _kyber.crypto_kem_keypair

def crypto_kem_enc(ct, ss, pk):
    return _kyber.crypto_kem_enc(ct, ss, pk)
crypto_kem_enc = _kyber.crypto_kem_enc

def crypto_kem_dec(ss, ct, sk):
    return _kyber.crypto_kem_dec(ss, ct, sk)
crypto_kem_dec = _kyber.crypto_kem_dec

def randombytes(x, xlen):
    return _kyber.randombytes(x, xlen)
randombytes = _kyber.randombytes
class Kyber(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Kyber, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Kyber, name)
    __repr__ = _swig_repr

    def __init__(self, *args):
        this = _kyber.new_Kyber(*args)
        try:
            self.this.append(this)
        except Exception:
            self.this = this
    __swig_destroy__ = _kyber.delete_Kyber
    __del__ = lambda self: None

    def getPK(self):
        return _kyber.Kyber_getPK(self)

    def getSK(self):
        return _kyber.Kyber_getSK(self)

    def getMyKey(self):
        return _kyber.Kyber_getMyKey(self)

    def getCypherText(self):
        return _kyber.Kyber_getCypherText(self)

    def kem_encode(self, other_pk):
        return _kyber.Kyber_kem_encode(self, other_pk)

    def kem_decode(self, cyphertext):
        return _kyber.Kyber_kem_decode(self, cyphertext)
Kyber_swigregister = _kyber.Kyber_swigregister
Kyber_swigregister(Kyber)

# This file is compatible with both classic and new-style classes.





# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from __future__ import print_function
from unittest import TestCase

from pyqrllib import pyqrllib


class TestXmssPool(TestCase):
    # sha2_input1 = 'hello'
    # sha2_expected_result1 = '2cf24dba5fb0a30e26e83b2ac5b9e29e1b161e5c1fa7425e73043362938b9824'
    # sha2_input2 = 'hello-qrl'
    # sha2_expected_result2 = '4ad6ad6c9ee6d2e52ebe4d635aa04052b7014e5e2e6b0de36da7648fac147703'

    def __init__(self, *args, **kwargs):
        super(TestXmssPool, self).__init__(*args, **kwargs)

    def test_pool(self):
        baseseed = pyqrllib.ucharVector(48, 0)
        pool = pyqrllib.XmssPool(baseseed, 6, 0, 3)
        self.assertFalse(pool.isAvailable())
        self.assertEqual(pool.getCurrentIndex(), 0)

        xmss = pool.getNextTree()
        self.assertEqual("0103002dcc3803df4475334b29eaa2516d1a9b36bc19eed0542cfbb501bf8de95d939b"
                         "25510d9876c7845b4694441bdc0e2be51f3d3f87f0c7775893845f25d49f9ef1",
                         pyqrllib.bin2hstr(xmss.getPK()))

        xmss = pool.getNextTree()
        self.assertEqual("010300be9caeafe11fa52edf722063c18616d6d0c6c30dba3e2c9369a6d9260f76818b"
                         "c424a1b6db5f26ef01ffa4aac8a08440d6a569bc56180b06f51b6ff6e4cc1b2e",
                         pyqrllib.bin2hstr(xmss.getPK()))

        xmss = pool.getNextTree()
        self.assertEqual("0103005deb91b1d311ecc8c6954e22f3e140ff3e6c04e40cad50c940e60abba3cf766a"
                         "5db9f39f58e532bea6765e4530cb581db1d6afbb7c05da3261ca4db21177afc5",
                         pyqrllib.bin2hstr(xmss.getPK()))

        xmss = pool.getNextTree()
        self.assertEqual("010300ea15c686e7b9691b8bac52ee4d0ed33bd75ec600f55b4476b857858460c2c983"
                         "90712040a5e03bca7a46715a90270ac2f8db8694ffa943091edb9018fa1dda04",
                         pyqrllib.bin2hstr(xmss.getPK()))



# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.


from __future__ import print_function

from unittest import TestCase

from pyqrllib import pyqrllib


class TestHelper(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestHelper, self).__init__(*args, **kwargs)

    def test_empty(self):
        self.assertFalse(pyqrllib.QRLHelper.addressIsValid(b''))



# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from __future__ import print_function

import unittest
from unittest import TestCase

from pyqrllib.kyber import Kyber


class TestKyber(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestKyber, self).__init__(*args, **kwargs)

    def test_exchange_keys(self):
        alice = Kyber()
        bob = Kyber()

        # Alice sends her public key to Bob
        alice_public_key = alice.getPK()

        # Bob receives the public key, derives a secret and a response
        bob.kem_encode(alice_public_key)
        cypherText = bob.getCypherText()

        # Bob sends the cyphertext to alice
        valid = alice.kem_decode(cypherText)

        # Now Alice and Bob share the same key
        alice_key = alice.getMyKey()
        bob_key = bob.getMyKey()

        self.assertTrue(valid)
        self.assertEqual(alice_key, bob_key)


if __name__ == '__main__':
    unittest.main()



# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.


from __future__ import print_function
from unittest import TestCase

from pyqrllib import pyqrllib
from pyqrllib.pyqrllib import mnemonic2bin, bin2mnemonic


class TestMisc(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestMisc, self).__init__(*args, **kwargs)

    def test_data_to_hex1(self):
        hexstring = pyqrllib.bin2hstr(b'\x00\x11\x22\x33', 0)
        self.assertEqual(hexstring, '00112233')

    def test_data_to_hex2(self):
        hexstring = pyqrllib.bin2hstr('test', 0)
        self.assertEqual(hexstring, '74657374')

    def test_mnemonic_words_odd_1(self):
        with self.assertRaises(ValueError):
            mnemonic2bin('absorb')

    def test_mnemonic_words_odd_2(self):
        with self.assertRaises(ValueError):
            mnemonic2bin('absorb bunny bunny')

    def test_mnemonic1(self):
        result = mnemonic2bin('aback absorb')
        self.assertEqual(tuple([0, 0, 16]), result)

    def test_mnemonic2(self):
        result = mnemonic2bin('absorb absorb')
        self.assertEqual(tuple([1, 0, 16]), result)

    def test_mnemonic3(self):
        long_mnemonic = "law bruise screen lunar than loft but franc strike asleep dwarf tavern dragon alarm " + \
                        "snack queen meadow thing far cotton add emblem strive probe zurich edge peer alight " + \
                        "libel won corn medal"
        exp_result = '7ad1e6c1083de2081221056dd8b0c142cdfa3fd053cd4ae288ee324cd30e027462d8eaaffff445a1105b7e4fc1302894'

        self.assertEqual(exp_result, pyqrllib.bin2hstr(mnemonic2bin(long_mnemonic)))

    def test_mnemonic4(self):
        bin = mnemonic2bin('absorb absorb')
        tmp_mnemonic = bin2mnemonic(bin)
        self.assertEqual('absorb absorb', tmp_mnemonic)

    def test_exception(self):
        with self.assertRaises(ValueError):
            pyqrllib.hstr2bin('Z')

        with self.assertRaises(ValueError):
            pyqrllib.hstr2bin('Z0')



from math import *
import functools

log_infinity = 9999

def delta_BKZ(b):
    """ The root hermite factor delta of BKZ-b
    """
    return ((pi*b)**(1./b) * b / (2*pi*exp(1)))**(1./(2.*b-2.))


def svp_plausible(b):
    """ log_2 of best plausible Quantum Cost of SVP in dimension b
    """
    return b *log(sqrt(4./3))/log(2)   # .2075 * b 


def svp_quantum(b):
    """ log_2 of best plausible Quantum Cost of SVP in dimension b
    """
    return b *log(sqrt(13./9))/log(2)   # .265 * b  [Laarhoven Thesis]


def svp_classical(b):
    """ log_2 of best known Quantum Cost of SVP in dimension b
    """
    return b *log(sqrt(3./2))/log(2)    # .292 * b [Becker Ducas Laarhoven Gama]


def nvec_sieve(b):
    """ Number of short vectors outputted by a sieve step of blocksize b
    """
    return b *log(sqrt(4./3))/log(2)    # .2075 * b


## Adding Memoization to this really slow function
# @functools.lru_cache(maxsize=2**20)
def construct_BKZ_shape(q, nq, n1, b):
    """ Simulate the (log) shape of a basis after the reduction of 
        a [q ... q, 1 ... 1] shape after BKZ-b reduction (nq many q's, n1 many 1's)
        This is implemented by constructing a longer shape and looking
        for the subshape with the right volume. Also outputs the index of the
        first vector <q, and the last >q.

        # Note: this implentation takes O(n). It is possible to output
        # a compressed description of the shape in time O(1), but it is much
        # more prone to making mistakes

    """ 
    d = nq+n1
    if b==0:
        L = nq*[log(q)] + n1*[0]
        return (nq, nq, L)


    slope = -2 * log(delta_BKZ(b))
    lq = log(q)
    B = int(floor(log(q) / - slope))    # Number of vectors in the sloppy region
    L = nq*[log(q)] + [lq + i * slope for i in range(1, B+1)] + n1*[0]

    x = 0
    lv = sum (L[:d])
    glv = nq*lq                     # Goal log volume

    while lv > glv:                 # While the current volume exceeeds goal volume, slide the window to the right
        lv -= L[x]
        lv += L[x+d]
        x += 1

    assert x <= B                   # Sanity check that we have not gone too far

    L = L[x:x+d]
    a = max(0, nq - x)             # The length of the [q, ... q] sequence
    B = min(B, d - a)              # The length of the GSA sequence

    diff = glv - lv
    assert abs(diff) < lq               # Sanity check the volume, up to the discretness of index error
    for i in range(a, a+B):        # Small shift of the GSA sequence to equiliBrate volume
        L[i] -= diff / B
    lv = sum(L) 
    assert lv/glv - 1 < 1e-6        # Sanity check the volume

    return (a, a + B, L)


def construct_BKZ_shape_randomized(q, nq, n1, b):
    """ Simulate the (log) shape of a basis after the reduction of 
        a [q ... q, 1 ... 1] shape after a randomization and a BKZ-b reduction 
        (such that no GS vectors gets smaller than 1)
    """ 
    glv = nq * log(q)
    d = nq+n1
    L = []

    slope = -2 * log(delta_BKZ(b))
    li = 0
    lv = 0
    for i in range(d):
        li -= slope
        lv += li
        if lv>glv:
            break
        L = [li]+L
    B = len(L)                  # The length of the sloppy sequence
    L += (d-B)*[0]
    a = 0                        # The length of the [q, ... q] sequence

    lv = sum(L) 
    diff = lv - glv
    #print diff, li
    #assert abs(diff) < li          # Sanity check the volume, up to the discretness of index error
    for i in range(a, a+B):        # Small shift of the GSA sequence to equiliBrate volume
        L[i] -= diff / B
    lv = sum(L) 
    assert abs(lv/glv - 1) < 1e-6        # Sanity check the volume

    return (a, a + B, L)




def BKZ_first_length(q, nq, n1, b):
    """ Simulate the length of the shortest expected vector in the first b-block
        after a BKZ-b reduction, ignoring the q-vector if some remains.
    """

    (_, _, L) = construct_BKZ_shape(q, nq, n1, b)
    rV = exp(sum(L[:b]) / b)                # Compute the root-volume of the first block
    delta = delta_BKZ(b)
    l = rV * delta**b
    return l

def BKZ_last_block_length(q, nq, n1, b):
    """ Simulate the length of the expected Gram-Schmidt vector at position d-b (d = n+m)
        after a BKZ-b reduction.
    """

    (_, _, L) = construct_BKZ_shape(q, nq, n1, b)
    return exp(L[nq + n1 - b])




from math import factorial as fac
from math import log, ceil, erf, sqrt


def gaussian_center_weight(sigma, t):
    """ Weight of the gaussian of std deviation s, on the interval [-t, t]
    :param x: (float)
    :param y: (float)
    :returns: erf( t / (sigma*\sqrt 2) )
    """
    return erf(t / (sigma * sqrt(2.)))


def binomial(x, y):
    """ Binomial coefficient
    :param x: (integer)
    :param y: (integer)
    :returns: y choose x
    """    
    try:
        binom = fac(x) // fac(y) // fac(x - y)
    except ValueError:
        binom = 0
    return binom


def centered_binomial_pdf(k, x):
    """ Probability density function of the centered binomial law of param k at x
    :param k: (integer)
    :param x: (integer)
    :returns: p_k(x)
    """    
    return binomial(2*k, x+k) / 2.**(2*k)


def build_centered_binomial_law(k):
    """ Construct the binomial law as a dictionnary
    :param k: (integer)
    :param x: (integer)
    :returns: A dictionnary {x:p_k(x) for x in {-k..k}}
    """    
    D = {}
    for i in range(-k, k+1):
        D[i] = centered_binomial_pdf(k, i)
    return D


def mod_switch(x, q, rq):
    """ Modulus switching (rounding to a different discretization of the Torus)
    :param x: value to round (integer)
    :param q: input modulus (integer)
    :param rq: output modulus (integer)
    """    
    return int(round(1.* rq * x / q) % rq)


def mod_centered(x, q):
    """ reduction mod q, centered (ie represented in -q/2 .. q/2)
    :param x: value to round (integer)
    :param q: input modulus (integer)
    """        
    a = x % q
    if a < q/2:
        return a
    return a - q


def build_mod_switching_error_law(q, rq):
    """ Construct Error law: law of the difference introduced by switching from and back a uniform value mod q
    :param q: original modulus (integer)
    :param rq: intermediate modulus (integer)
    """        
    D = {}
    V = {}
    for x in range(q):
        y = mod_switch(x, q, rq)
        z = mod_switch(y, rq, q)
        d = mod_centered(x - z, q)
        D[d] = D.get(d, 0) + 1./q
        V[y] = V.get(y, 0) + 1

    return D


def law_convolution(A, B):
    """ Construct the convolution of two laws (sum of independent variables from two input laws)
    :param A: first input law (dictionnary)
    :param B: second input law (dictionnary)
    """        

    C = {}
    for a in A:
        for b in B:
            c = a+b
            C[c] = C.get(c, 0) + A[a] * B[b]
    return C


def law_product(A, B):
    """ Construct the law of the product of independent variables from two input laws
    :param A: first input law (dictionnary)
    :param B: second input law (dictionnary)
    """        
    C = {}
    for a in A:
        for b in B:
            c = a*b
            C[c] = C.get(c, 0) + A[a] * B[b]
    return C


def clean_dist(A):
    """ Clean a distribution to accelerate further computation (drop element of the support with proba less than 2^-300)
    :param A: input law (dictionnary)
    """        
    B = {}
    for (x, y) in A.items():
        if y>2**(-300):
            B[x] = y
    return B


def iter_law_convolution(A, i):
    """ compute the -ith forld convolution of a distribution (using double-and-add)
    :param A: first input law (dictionnary)
    :param i: (integer)
    """    
    D = {0: 1.0}
    i_bin = bin(i)[2:]  # binary representation of n
    for ch in i_bin:
        D = law_convolution(D, D)
        D = clean_dist(D)
        if ch == '1':
            D = law_convolution(D, A)
            D = clean_dist(D)
    return D


def tail_probability(D, t):
    '''
    Probability that an drawn from D is strictly greater than t in absolute value
    :param D: Law (Dictionnary)
    :param t: tail parameter (integer)
    '''
    s = 0
    ma = max(D.keys())
    if t >= ma:
        return 0
    for i in reversed(range(int(ceil(t)), ma)):  # Summing in reverse for better numerical precision (assuming tails are decreasing)
        s += D.get(i, 0) + D.get(-i, 0)
    return s



# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from __future__ import print_function

from time import time
from unittest import TestCase

import pytest

from pyqrllib import pyqrllib


class TestXmssBasic(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestXmssBasic, self).__init__(*args, **kwargs)

    def test_xmss_creation_height4(self):
        HEIGHT = 4
        seed = pyqrllib.ucharVector(48, 0)
        xmss = pyqrllib.XmssBasic(seed, HEIGHT, pyqrllib.SHAKE_128, pyqrllib.SHA256_2X)

        expected_address = "01020095f03f084bcb29b96b0529c17ce92c54c1e8290193a93803812ead95e8e6902506b67897"
        expected_PK = "010200c25188b585f731c128e2b457069e" \
                      "afd1e3fa3961605af8c58a1aec4d82ac" \
                      "316d3191da3442686282b3d5160f25cf" \
                      "162a517fd2131f83fbf2698a58f9c46a" \
                      "fc5d"

        self.assertEqual(expected_PK, pyqrllib.bin2hstr(xmss.getPK()))
        self.assertEqual(expected_address, pyqrllib.bin2hstr(xmss.getAddress()))

        tmp_addr = pyqrllib.QRLHelper.getAddress(xmss.getPK())
        self.assertEqual(expected_address, pyqrllib.bin2hstr(tmp_addr))

        descr = pyqrllib.QRLDescriptor.fromExtendedPK(xmss.getPK())
        self.assertEqual(4, descr.getHeight())
        self.assertEqual(pyqrllib.SHAKE_128, descr.getHashFunction())

    def test_xmss_creation_height6(self):
        HEIGHT = 6
        seed = pyqrllib.ucharVector(48, 0)
        xmss = pyqrllib.XmssBasic(seed, HEIGHT, pyqrllib.SHAKE_128, pyqrllib.SHA256_2X)

        expected_address = "0103008b0e18dd0bac2c3fdc9a48e10fc466eef899ef074449d12ddf050317b2083527aee74bc3"

        expected_PK = "010300859060f15adc3825adeec85c7483" \
                      "d868e898bc5117d0cff04ab1343916d4" \
                      "07af3191da3442686282b3d5160f25cf" \
                      "162a517fd2131f83fbf2698a58f9c46a" \
                      "fc5d"

        self.assertEqual(expected_PK, pyqrllib.bin2hstr(xmss.getPK()))
        self.assertEqual(expected_address, pyqrllib.bin2hstr(xmss.getAddress()))

        tmp_addr = pyqrllib.QRLHelper.getAddress(xmss.getPK())
        self.assertEqual(expected_address, pyqrllib.bin2hstr(tmp_addr))

        descr = pyqrllib.QRLDescriptor.fromExtendedPK(xmss.getPK())
        self.assertEqual(6, descr.getHeight())
        self.assertEqual(pyqrllib.SHAKE_128, descr.getHashFunction())

    def test_xmss(self):
        HEIGHT = 6

        seed = pyqrllib.ucharVector(48, 0)
        xmss = pyqrllib.XmssBasic(seed, HEIGHT, pyqrllib.SHAKE_128, pyqrllib.SHA256_2X)

        # print("Seed", len(seed))
        # print(pyqrllib.bin2hstr(seed, 48))
        #
        # print("PK  ", len(xmss.getPK()))
        # print(pyqrllib.bin2hstr(xmss.getPK(), 48))
        #
        # print("SK  ", len(xmss.getSK()))
        # print(pyqrllib.bin2hstr(xmss.getSK(), 48))

        self.assertIsNotNone(xmss)
        self.assertEqual(xmss.getHeight(), HEIGHT)

        message = pyqrllib.ucharVector([i for i in range(32)])
        # print("Msg ", len(message))
        # print(pyqrllib.bin2hstr(message, 48))

        # Sign message
        signature = bytearray(xmss.sign(message))

        # print("Sig ", len(signature))
        # print(pyqrllib.bin2hstr(signature, 128))
        #
        # print('----------------------------------------------------------------------')
        # Verify signature
        start = time()
        for i in range(1000):
            self.assertTrue(pyqrllib.XmssBasic.verify(message,
                                                      signature,
                                                      xmss.getPK()))
        end = time()
        print(end - start)

        # Touch the signature
        signature[100] += 1
        self.assertFalse(pyqrllib.XmssBasic.verify(message,
                                                   signature,
                                                   xmss.getPK()))
        signature[100] -= 1
        self.assertTrue(pyqrllib.XmssBasic.verify(message,
                                                  signature,
                                                  xmss.getPK()))

        # Touch the message
        message[2] += 1
        self.assertFalse(pyqrllib.XmssBasic.verify(message,
                                                   signature,
                                                   xmss.getPK()))
        message[2] -= 1
        self.assertTrue(pyqrllib.XmssBasic.verify(message,
                                                  signature,
                                                  xmss.getPK()))

    def test_xmss_exception_constructor(self):
        HEIGHT = 7
        seed = pyqrllib.ucharVector(48, 0)

        with pytest.raises(ValueError):
            pyqrllib.XmssFast(seed, HEIGHT, pyqrllib.SHAKE_128)

    def test_xmss_exception_verify(self):
        message = pyqrllib.ucharVector(48, 0)
        signature = pyqrllib.ucharVector(2287, 0)
        pk = pyqrllib.ucharVector(67, 0)

        self.assertFalse(pyqrllib.XmssFast.verify(message, signature, pk))

    def test_xmss_change_index_too_high(self):
        HEIGHT = 4
        seed = pyqrllib.ucharVector(48, 0)
        xmss = pyqrllib.XmssFast(seed, HEIGHT, pyqrllib.SHAKE_128)

        with pytest.raises(ValueError):
            xmss.setIndex(20)

    def test_xmss_change_index_high(self):
        HEIGHT = 4
        seed = pyqrllib.ucharVector(48, 0)
        xmss = pyqrllib.XmssFast(seed, HEIGHT, pyqrllib.SHAKE_128)

        with pytest.raises(ValueError):
            xmss.setIndex(16)

    def test_xmss_change_index_limit(self):
        HEIGHT = 4
        seed = pyqrllib.ucharVector(48, 0)
        xmss = pyqrllib.XmssFast(seed, HEIGHT, pyqrllib.SHAKE_128)

        xmss.setIndex(15)
        self.assertEqual(15, xmss.getIndex())

    def test_xmss_change_index(self):
        HEIGHT = 4
        seed = pyqrllib.ucharVector(48, 0)
        xmss = pyqrllib.XmssFast(seed, HEIGHT, pyqrllib.SHAKE_128)

        xmss.setIndex(0)
        self.assertEqual(0, xmss.getIndex())



# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from __future__ import print_function
from unittest import TestCase

from pyqrllib import pyqrllib


class TestSha2_256(TestCase):
    sha2_input1 = 'hello'
    sha2_expected_result1 = '2cf24dba5fb0a30e26e83b2ac5b9e29e1b161e5c1fa7425e73043362938b9824'
    sha2_input2 = 'hello-qrl'
    sha2_expected_result2 = '4ad6ad6c9ee6d2e52ebe4d635aa04052b7014e5e2e6b0de36da7648fac147703'

    def __init__(self, *args, **kwargs):
        super(TestSha2_256, self).__init__(*args, **kwargs)

    def check_sha_result(self, data_text, expected):
        hex_in_before = pyqrllib.bin2hstr(pyqrllib.str2bin(data_text))
        data_out = pyqrllib.sha2_256(pyqrllib.str2bin(data_text))

        # This is just to keep as an example. Things could be compared without converting to hex
        hex_in = pyqrllib.bin2hstr(pyqrllib.str2bin(data_text))
        hex_out = pyqrllib.bin2hstr(data_out)

        self.assertEqual(hex_in_before, hex_in)
        self.assertEqual(expected, hex_out)

    def check_sha_n_result(self, data_text, expected, count):
        hex_in_before = pyqrllib.bin2hstr(pyqrllib.str2bin(data_text))
        data_out = pyqrllib.sha2_256_n(pyqrllib.str2bin(data_text), count)

        # This is just to keep as an example. Things could be compared without converting to hex
        hex_in = pyqrllib.bin2hstr(pyqrllib.str2bin(data_text))
        hex_out = pyqrllib.bin2hstr(data_out)

        self.assertEqual(hex_in_before, hex_in)
        self.assertEqual(expected, hex_out)

    def test_check_sha2_256(self):
        self.check_sha_result(self.sha2_input1, self.sha2_expected_result1)
        self.check_sha_result(self.sha2_input2, self.sha2_expected_result2)

    def test_sha_n_result(self):
        self.check_sha_n_result("This is a test X",
                                "a11609b2cc5f26619fcc865473246c9ac59861383a3c4edd2433230258afa03b", 1)
        self.check_sha_n_result("This is a test X",
                                "3be2d7e048d22de2c117465e5b4b819e764352680027c9790a53a7326d62a0fe", 16)



import operator as op
from math import factorial as fac
from math import sqrt, log
import sys
from proba_util import *

def p2_cyclotomic_final_error_distribution(ps):
    """ construct the final error distribution in our encryption scheme
    :param ps: parameter set (ParameterSet)
    """
    chis = build_centered_binomial_law(ps.ks)           # LWE error law for the key
    chie = build_centered_binomial_law(ps.ke)           # LWE error law for the ciphertext 
    Rk = build_mod_switching_error_law(ps.q, ps.rqk)    # Rounding error public key
    Rc = build_mod_switching_error_law(ps.q, ps.rqc)    # rounding error first ciphertext
    chiRs = law_convolution(chis, Rk)                   # LWE+Rounding error key
    chiRe = law_convolution(chie, Rc)                   # LWE + rounding error ciphertext

    B1 = law_product(chie, chiRs)                       # (LWE+Rounding error) * LWE (as in a E*S product)
    B2 = law_product(chis, chiRe)
  
    C1 = iter_law_convolution(B1, ps.m * ps.n)
    C2 = iter_law_convolution(B2, ps.m * ps.n)
    
    C=law_convolution(C1, C2)
    
    R2 = build_mod_switching_error_law(ps.q, ps.rq2)    # Rounding2 (in the ciphertext mask part)
    F = law_convolution(R2, chie)                       # LWE+Rounding2 error
    D = law_convolution(C, F)                           # Final error
    return D


def p2_cyclotomic_error_probability(ps):
    F = p2_cyclotomic_final_error_distribution(ps)
    proba = tail_probability(F, ps.q/4)
    return F, ps.n*proba


#!/usr/bin/env python
# -*- coding: utf-8 -*-
import sys
import os
import subprocess

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from distutils.sysconfig import get_python_inc
import distutils.sysconfig as sysconfig


class CMakeBuild(build_ext):
    def run(self):
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)

        env = os.environ.copy()
        env['CXXFLAGS'] = env.get('CXXFLAGS', '')
        env['CXXFLAGS'] += ' -DVERSION_INFO=\\"' + self.distribution.get_version() + '\\"'

        for ext in self.extensions:
            extension_path = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))

            cmake_call = ['cmake', ext.sourcedir,
                          '-DBUILD_PYTHON=ON',
                          '-DBUILD_TESTS=OFF',
                          '-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extension_path,
                          '-DCMAKE_BUILD_TYPE=Release']

            # Detect conda
            if sys.platform == 'darwin' and 'CONDA_DEFAULT_ENV' in os.environ:
                print('OSX + Conda environment detected')
                python_include_dir = get_python_inc()
                python_library = os.path.join(sysconfig.get_config_var('LIBDIR'), sysconfig.get_config_var('LDLIBRARY'))
                cmake_call.extend(['-DPYTHON_INCLUDE_DIR=' + python_include_dir,
                                   '-DPYTHON_LIBRARY=' + python_library])

            subprocess.check_call(cmake_call, cwd=self.build_temp, env=env)

            subprocess.check_call(['cmake', '--build', '.',
                                   '--config', 'Release', '--', '-j2'], cwd=self.build_temp)


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir='', *args, **kw):
        Extension.__init__(self, name, sources=[], *args, **kw)
        self.sourcedir = os.path.abspath(sourcedir)


def setup_package():
    needs_sphinx = {'build_sphinx', 'upload_docs'}.intersection(sys.argv)
    sphinx = ['sphinx'] if needs_sphinx else []
    cmake = []

    # noinspection PyInterpreter
    setup(setup_requires=['six', 'pytest-runner', 'pyscaffold>3'] + sphinx + cmake,
          packages=['pyqrllib', ],
          tests_require=['pytest', 'pytest-cov'],
          ext_modules=[CMakeExtension('pyqrllib')],
          cmdclass=dict(build_ext=CMakeBuild),
          use_pyscaffold=True)


if __name__ == "__main__":
    setup_package()



# This file was automatically generated by SWIG (http://www.swig.org).
# Version 3.0.8
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.





from sys import version_info
if version_info >= (2, 6, 0):
    def swig_import_helper():
        from os.path import dirname
        import imp
        fp = None
        try:
            fp, pathname, description = imp.find_module('_pyqrllib', [dirname(__file__)])
        except ImportError:
            import _pyqrllib
            return _pyqrllib
        if fp is not None:
            try:
                _mod = imp.load_module('_pyqrllib', fp, pathname, description)
            finally:
                fp.close()
            return _mod
    _pyqrllib = swig_import_helper()
    del swig_import_helper
else:
    import _pyqrllib
del version_info
try:
    _swig_property = property
except NameError:
    pass  # Python < 2.2 doesn't have 'property'.


def _swig_setattr_nondynamic(self, class_type, name, value, static=1):
    if (name == "thisown"):
        return self.this.own(value)
    if (name == "this"):
        if type(value).__name__ == 'SwigPyObject':
            self.__dict__[name] = value
            return
    method = class_type.__swig_setmethods__.get(name, None)
    if method:
        return method(self, value)
    if (not static):
        if _newclass:
            object.__setattr__(self, name, value)
        else:
            self.__dict__[name] = value
    else:
        raise AttributeError("You cannot add attributes to %s" % self)


def _swig_setattr(self, class_type, name, value):
    return _swig_setattr_nondynamic(self, class_type, name, value, 0)


def _swig_getattr_nondynamic(self, class_type, name, static=1):
    if (name == "thisown"):
        return self.this.own()
    method = class_type.__swig_getmethods__.get(name, None)
    if method:
        return method(self)
    if (not static):
        return object.__getattr__(self, name)
    else:
        raise AttributeError(name)

def _swig_getattr(self, class_type, name):
    return _swig_getattr_nondynamic(self, class_type, name, 0)


def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)

try:
    _object = object
    _newclass = 1
except AttributeError:
    class _object:
        pass
    _newclass = 0



def cdata(ptr, nelements=1):
    return _pyqrllib.cdata(ptr, nelements)
cdata = _pyqrllib.cdata

def memmove(data, indata):
    return _pyqrllib.memmove(data, indata)
memmove = _pyqrllib.memmove
class SwigPyIterator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, SwigPyIterator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, SwigPyIterator, name)

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined - class is abstract")
    __repr__ = _swig_repr
    __swig_destroy__ = _pyqrllib.delete_SwigPyIterator
    __del__ = lambda self: None

    def value(self):
        return _pyqrllib.SwigPyIterator_value(self)

    def incr(self, n=1):
        return _pyqrllib.SwigPyIterator_incr(self, n)

    def decr(self, n=1):
        return _pyqrllib.SwigPyIterator_decr(self, n)

    def distance(self, x):
        return _pyqrllib.SwigPyIterator_distance(self, x)

    def equal(self, x):
        return _pyqrllib.SwigPyIterator_equal(self, x)

    def copy(self):
        return _pyqrllib.SwigPyIterator_copy(self)

    def next(self):
        return _pyqrllib.SwigPyIterator_next(self)

    def __next__(self):
        return _pyqrllib.SwigPyIterator___next__(self)

    def previous(self):
        return _pyqrllib.SwigPyIterator_previous(self)

    def advance(self, n):
        return _pyqrllib.SwigPyIterator_advance(self, n)

    def __eq__(self, x):
        return _pyqrllib.SwigPyIterator___eq__(self, x)

    def __ne__(self, x):
        return _pyqrllib.SwigPyIterator___ne__(self, x)

    def __iadd__(self, n):
        return _pyqrllib.SwigPyIterator___iadd__(self, n)

    def __isub__(self, n):
        return _pyqrllib.SwigPyIterator___isub__(self, n)

    def __add__(self, n):
        return _pyqrllib.SwigPyIterator___add__(self, n)

    def __sub__(self, *args):
        return _pyqrllib.SwigPyIterator___sub__(self, *args)
    def __iter__(self):
        return self
SwigPyIterator_swigregister = _pyqrllib.SwigPyIterator_swigregister
SwigPyIterator_swigregister(SwigPyIterator)


_pyqrllib.SHARED_PTR_DISOWN_swigconstant(_pyqrllib)
SHARED_PTR_DISOWN = _pyqrllib.SHARED_PTR_DISOWN
class ucharCArray(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, ucharCArray, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, ucharCArray, name)
    __repr__ = _swig_repr

    def __init__(self, nelements):
        this = _pyqrllib.new_ucharCArray(nelements)
        try:
            self.this.append(this)
        except Exception:
            self.this = this
    __swig_destroy__ = _pyqrllib.delete_ucharCArray
    __del__ = lambda self: None

    def __getitem__(self, index):
        return _pyqrllib.ucharCArray___getitem__(self, index)

    def __setitem__(self, index, value):
        return _pyqrllib.ucharCArray___setitem__(self, index, value)

    def cast(self):
        return _pyqrllib.ucharCArray_cast(self)
    __swig_getmethods__["frompointer"] = lambda x: _pyqrllib.ucharCArray_frompointer
    if _newclass:
        frompointer = staticmethod(_pyqrllib.ucharCArray_frompointer)
ucharCArray_swigregister = _pyqrllib.ucharCArray_swigregister
ucharCArray_swigregister(ucharCArray)

def ucharCArray_frompointer(t):
    return _pyqrllib.ucharCArray_frompointer(t)
ucharCArray_frompointer = _pyqrllib.ucharCArray_frompointer

class uintCArray(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, uintCArray, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, uintCArray, name)
    __repr__ = _swig_repr

    def __init__(self, nelements):
        this = _pyqrllib.new_uintCArray(nelements)
        try:
            self.this.append(this)
        except Exception:
            self.this = this
    __swig_destroy__ = _pyqrllib.delete_uintCArray
    __del__ = lambda self: None

    def __getitem__(self, index):
        return _pyqrllib.uintCArray___getitem__(self, index)

    def __setitem__(self, index, value):
        return _pyqrllib.uintCArray___setitem__(self, index, value)

    def cast(self):
        return _pyqrllib.uintCArray_cast(self)
    __swig_getmethods__["frompointer"] = lambda x: _pyqrllib.uintCArray_frompointer
    if _newclass:
        frompointer = staticmethod(_pyqrllib.uintCArray_frompointer)
uintCArray_swigregister = _pyqrllib.uintCArray_swigregister
uintCArray_swigregister(uintCArray)

def uintCArray_frompointer(t):
    return _pyqrllib.uintCArray_frompointer(t)
uintCArray_frompointer = _pyqrllib.uintCArray_frompointer

class uint32CArray(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, uint32CArray, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, uint32CArray, name)
    __repr__ = _swig_repr

    def __init__(self, nelements):
        this = _pyqrllib.new_uint32CArray(nelements)
        try:
            self.this.append(this)
        except Exception:
            self.this = this
    __swig_destroy__ = _pyqrllib.delete_uint32CArray
    __del__ = lambda self: None

    def __getitem__(self, index):
        return _pyqrllib.uint32CArray___getitem__(self, index)

    def __setitem__(self, index, value):
        return _pyqrllib.uint32CArray___setitem__(self, index, value)

    def cast(self):
        return _pyqrllib.uint32CArray_cast(self)
    __swig_getmethods__["frompointer"] = lambda x: _pyqrllib.uint32CArray_frompointer
    if _newclass:
        frompointer = staticmethod(_pyqrllib.uint32CArray_frompointer)
uint32CArray_swigregister = _pyqrllib.uint32CArray_swigregister
uint32CArray_swigregister(uint32CArray)

def uint32CArray_frompointer(t):
    return _pyqrllib.uint32CArray_frompointer(t)
uint32CArray_frompointer = _pyqrllib.uint32CArray_frompointer

class intVector(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, intVector, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, intVector, name)
    __repr__ = _swig_repr

    def iterator(self):
        return _pyqrllib.intVector_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self):
        return _pyqrllib.intVector___nonzero__(self)

    def __bool__(self):
        return _pyqrllib.intVector___bool__(self)

    def __len__(self):
        return _pyqrllib.intVector___len__(self)

    def __getslice__(self, i, j):
        return _pyqrllib.intVector___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _pyqrllib.intVector___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _pyqrllib.intVector___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _pyqrllib.intVector___delitem__(self, *args)

    def __getitem__(self, *args):
        return _pyqrllib.intVector___getitem__(self, *args)

    def __setitem__(self, *args):
        return _pyqrllib.intVector___setitem__(self, *args)

    def pop(self):
        return _pyqrllib.intVector_pop(self)

    def append(self, x):
        return _pyqrllib.intVector_append(self, x)

    def empty(self):
        return _pyqrllib.intVector_empty(self)

    def size(self):
        return _pyqrllib.intVector_size(self)

    def swap(self, v):
        return _pyqrllib.intVector_swap(self, v)

    def begin(self):
        return _pyqrllib.intVector_begin(self)

    def end(self):
        return _pyqrllib.intVector_end(self)

    def rbegin(self):
        return _pyqrllib.intVector_rbegin(self)

    def rend(self):
        return _pyqrllib.intVector_rend(self)

    def clear(self):
        return _pyqrllib.intVector_clear(self)

    def get_allocator(self):
        return _pyqrllib.intVector_get_allocator(self)

    def pop_back(self):
        return _pyqrllib.intVector_pop_back(self)

    def erase(self, *args):
        return _pyqrllib.intVector_erase(self, *args)

    def __init__(self, *args):
        this = _pyqrllib.new_intVector(*args)
        try:
            self.this.append(this)
        except Exception:
            self.this = this

    def push_back(self, x):
        return _pyqrllib.intVector_push_back(self, x)

    def front(self):
        return _pyqrllib.intVector_front(self)

    def back(self):
        return _pyqrllib.intVector_back(self)

    def assign(self, n, x):
        return _pyqrllib.intVector_assign(self, n, x)

    def resize(self, *args):
        return _pyqrllib.intVector_resize(self, *args)

    def insert(self, *args):
        return _pyqrllib.intVector_insert(self, *args)

    def reserve(self, n):
        return _pyqrllib.intVector_reserve(self, n)

    def capacity(self):
        return _pyqrllib.intVector_capacity(self)
    __swig_destroy__ = _pyqrllib.delete_intVector
    __del__ = lambda self: None
intVector_swigregister = _pyqrllib.intVector_swigregister
intVector_swigregister(intVector)

class uintVector(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, uintVector, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, uintVector, name)
    __repr__ = _swig_repr

    def iterator(self):
        return _pyqrllib.uintVector_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self):
        return _pyqrllib.uintVector___nonzero__(self)

    def __bool__(self):
        return _pyqrllib.uintVector___bool__(self)

    def __len__(self):
        return _pyqrllib.uintVector___len__(self)

    def __getslice__(self, i, j):
        return _pyqrllib.uintVector___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _pyqrllib.uintVector___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _pyqrllib.uintVector___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _pyqrllib.uintVector___delitem__(self, *args)

    def __getitem__(self, *args):
        return _pyqrllib.uintVector___getitem__(self, *args)

    def __setitem__(self, *args):
        return _pyqrllib.uintVector___setitem__(self, *args)

    def pop(self):
        return _pyqrllib.uintVector_pop(self)

    def append(self, x):
        return _pyqrllib.uintVector_append(self, x)

    def empty(self):
        return _pyqrllib.uintVector_empty(self)

    def size(self):
        return _pyqrllib.uintVector_size(self)

    def swap(self, v):
        return _pyqrllib.uintVector_swap(self, v)

    def begin(self):
        return _pyqrllib.uintVector_begin(self)

    def end(self):
        return _pyqrllib.uintVector_end(self)

    def rbegin(self):
        return _pyqrllib.uintVector_rbegin(self)

    def rend(self):
        return _pyqrllib.uintVector_rend(self)

    def clear(self):
        return _pyqrllib.uintVector_clear(self)

    def get_allocator(self):
        return _pyqrllib.uintVector_get_allocator(self)

    def pop_back(self):
        return _pyqrllib.uintVector_pop_back(self)

    def erase(self, *args):
        return _pyqrllib.uintVector_erase(self, *args)

    def __init__(self, *args):
        this = _pyqrllib.new_uintVector(*args)
        try:
            self.this.append(this)
        except Exception:
            self.this = this

    def push_back(self, x):
        return _pyqrllib.uintVector_push_back(self, x)

    def front(self):
        return _pyqrllib.uintVector_front(self)

    def back(self):
        return _pyqrllib.uintVector_back(self)

    def assign(self, n, x):
        return _pyqrllib.uintVector_assign(self, n, x)

    def resize(self, *args):
        return _pyqrllib.uintVector_resize(self, *args)

    def insert(self, *args):
        return _pyqrllib.uintVector_insert(self, *args)

    def reserve(self, n):
        return _pyqrllib.uintVector_reserve(self, n)

    def capacity(self):
        return _pyqrllib.uintVector_capacity(self)
    __swig_destroy__ = _pyqrllib.delete_uintVector
    __del__ = lambda self: None
uintVector_swigregister = _pyqrllib.uintVector_swigregister
uintVector_swigregister(uintVector)

class ucharVector(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, ucharVector, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, ucharVector, name)
    __repr__ = _swig_repr

    def iterator(self):
        return _pyqrllib.ucharVector_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self):
        return _pyqrllib.ucharVector___nonzero__(self)

    def __bool__(self):
        return _pyqrllib.ucharVector___bool__(self)

    def __len__(self):
        return _pyqrllib.ucharVector___len__(self)

    def __getslice__(self, i, j):
        return _pyqrllib.ucharVector___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _pyqrllib.ucharVector___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _pyqrllib.ucharVector___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _pyqrllib.ucharVector___delitem__(self, *args)

    def __getitem__(self, *args):
        return _pyqrllib.ucharVector___getitem__(self, *args)

    def __setitem__(self, *args):
        return _pyqrllib.ucharVector___setitem__(self, *args)

    def pop(self):
        return _pyqrllib.ucharVector_pop(self)

    def append(self, x):
        return _pyqrllib.ucharVector_append(self, x)

    def empty(self):
        return _pyqrllib.ucharVector_empty(self)

    def size(self):
        return _pyqrllib.ucharVector_size(self)

    def swap(self, v):
        return _pyqrllib.ucharVector_swap(self, v)

    def begin(self):
        return _pyqrllib.ucharVector_begin(self)

    def end(self):
        return _pyqrllib.ucharVector_end(self)

    def rbegin(self):
        return _pyqrllib.ucharVector_rbegin(self)

    def rend(self):
        return _pyqrllib.ucharVector_rend(self)

    def clear(self):
        return _pyqrllib.ucharVector_clear(self)

    def get_allocator(self):
        return _pyqrllib.ucharVector_get_allocator(self)

    def pop_back(self):
        return _pyqrllib.ucharVector_pop_back(self)

    def erase(self, *args):
        return _pyqrllib.ucharVector_erase(self, *args)

    def __init__(self, *args):
        this = _pyqrllib.new_ucharVector(*args)
        try:
            self.this.append(this)
        except Exception:
            self.this = this

    def push_back(self, x):
        return _pyqrllib.ucharVector_push_back(self, x)

    def front(self):
        return _pyqrllib.ucharVector_front(self)

    def back(self):
        return _pyqrllib.ucharVector_back(self)

    def assign(self, n, x):
        return _pyqrllib.ucharVector_assign(self, n, x)

    def resize(self, *args):
        return _pyqrllib.ucharVector_resize(self, *args)

    def insert(self, *args):
        return _pyqrllib.ucharVector_insert(self, *args)

    def reserve(self, n):
        return _pyqrllib.ucharVector_reserve(self, n)

    def capacity(self):
        return _pyqrllib.ucharVector_capacity(self)
    __swig_destroy__ = _pyqrllib.delete_ucharVector
    __del__ = lambda self: None
ucharVector_swigregister = _pyqrllib.ucharVector_swigregister
ucharVector_swigregister(ucharVector)

class charVector(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, charVector, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, charVector, name)
    __repr__ = _swig_repr

    def iterator(self):
        return _pyqrllib.charVector_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self):
        return _pyqrllib.charVector___nonzero__(self)

    def __bool__(self):
        return _pyqrllib.charVector___bool__(self)

    def __len__(self):
        return _pyqrllib.charVector___len__(self)

    def __getslice__(self, i, j):
        return _pyqrllib.charVector___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _pyqrllib.charVector___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _pyqrllib.charVector___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _pyqrllib.charVector___delitem__(self, *args)

    def __getitem__(self, *args):
        return _pyqrllib.charVector___getitem__(self, *args)

    def __setitem__(self, *args):
        return _pyqrllib.charVector___setitem__(self, *args)

    def pop(self):
        return _pyqrllib.charVector_pop(self)

    def append(self, x):
        return _pyqrllib.charVector_append(self, x)

    def empty(self):
        return _pyqrllib.charVector_empty(self)

    def size(self):
        return _pyqrllib.charVector_size(self)

    def swap(self, v):
        return _pyqrllib.charVector_swap(self, v)

    def begin(self):
        return _pyqrllib.charVector_begin(self)

    def end(self):
        return _pyqrllib.charVector_end(self)

    def rbegin(self):
        return _pyqrllib.charVector_rbegin(self)

    def rend(self):
        return _pyqrllib.charVector_rend(self)

    def clear(self):
        return _pyqrllib.charVector_clear(self)

    def get_allocator(self):
        return _pyqrllib.charVector_get_allocator(self)

    def pop_back(self):
        return _pyqrllib.charVector_pop_back(self)

    def erase(self, *args):
        return _pyqrllib.charVector_erase(self, *args)

    def __init__(self, *args):
        this = _pyqrllib.new_charVector(*args)
        try:
            self.this.append(this)
        except Exception:
            self.this = this

    def push_back(self, x):
        return _pyqrllib.charVector_push_back(self, x)

    def front(self):
        return _pyqrllib.charVector_front(self)

    def back(self):
        return _pyqrllib.charVector_back(self)

    def assign(self, n, x):
        return _pyqrllib.charVector_assign(self, n, x)

    def resize(self, *args):
        return _pyqrllib.charVector_resize(self, *args)

    def insert(self, *args):
        return _pyqrllib.charVector_insert(self, *args)

    def reserve(self, n):
        return _pyqrllib.charVector_reserve(self, n)

    def capacity(self):
        return _pyqrllib.charVector_capacity(self)
    __swig_destroy__ = _pyqrllib.delete_charVector
    __del__ = lambda self: None
charVector_swigregister = _pyqrllib.charVector_swigregister
charVector_swigregister(charVector)

class doubleVector(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, doubleVector, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, doubleVector, name)
    __repr__ = _swig_repr

    def iterator(self):
        return _pyqrllib.doubleVector_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self):
        return _pyqrllib.doubleVector___nonzero__(self)

    def __bool__(self):
        return _pyqrllib.doubleVector___bool__(self)

    def __len__(self):
        return _pyqrllib.doubleVector___len__(self)

    def __getslice__(self, i, j):
        return _pyqrllib.doubleVector___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _pyqrllib.doubleVector___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _pyqrllib.doubleVector___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _pyqrllib.doubleVector___delitem__(self, *args)

    def __getitem__(self, *args):
        return _pyqrllib.doubleVector___getitem__(self, *args)

    def __setitem__(self, *args):
        return _pyqrllib.doubleVector___setitem__(self, *args)

    def pop(self):
        return _pyqrllib.doubleVector_pop(self)

    def append(self, x):
        return _pyqrllib.doubleVector_append(self, x)

    def empty(self):
        return _pyqrllib.doubleVector_empty(self)

    def size(self):
        return _pyqrllib.doubleVector_size(self)

    def swap(self, v):
        return _pyqrllib.doubleVector_swap(self, v)

    def begin(self):
        return _pyqrllib.doubleVector_begin(self)

    def end(self):
        return _pyqrllib.doubleVector_end(self)

    def rbegin(self):
        return _pyqrllib.doubleVector_rbegin(self)

    def rend(self):
        return _pyqrllib.doubleVector_rend(self)

    def clear(self):
        return _pyqrllib.doubleVector_clear(self)

    def get_allocator(self):
        return _pyqrllib.doubleVector_get_allocator(self)

    def pop_back(self):
        return _pyqrllib.doubleVector_pop_back(self)

    def erase(self, *args):
        return _pyqrllib.doubleVector_erase(self, *args)

    def __init__(self, *args):
        this = _pyqrllib.new_doubleVector(*args)
        try:
            self.this.append(this)
        except Exception:
            self.this = this

    def push_back(self, x):
        return _pyqrllib.doubleVector_push_back(self, x)

    def front(self):
        return _pyqrllib.doubleVector_front(self)

    def back(self):
        return _pyqrllib.doubleVector_back(self)

    def assign(self, n, x):
        return _pyqrllib.doubleVector_assign(self, n, x)

    def resize(self, *args):
        return _pyqrllib.doubleVector_resize(self, *args)

    def insert(self, *args):
        return _pyqrllib.doubleVector_insert(self, *args)

    def reserve(self, n):
        return _pyqrllib.doubleVector_reserve(self, n)

    def capacity(self):
        return _pyqrllib.doubleVector_capacity(self)
    __swig_destroy__ = _pyqrllib.delete_doubleVector
    __del__ = lambda self: None
doubleVector_swigregister = _pyqrllib.doubleVector_swigregister
doubleVector_swigregister(doubleVector)

class _string_list(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, _string_list, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, _string_list, name)
    __repr__ = _swig_repr

    def iterator(self):
        return _pyqrllib._string_list_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self):
        return _pyqrllib._string_list___nonzero__(self)

    def __bool__(self):
        return _pyqrllib._string_list___bool__(self)

    def __len__(self):
        return _pyqrllib._string_list___len__(self)

    def __getslice__(self, i, j):
        return _pyqrllib._string_list___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _pyqrllib._string_list___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _pyqrllib._string_list___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _pyqrllib._string_list___delitem__(self, *args)

    def __getitem__(self, *args):
        return _pyqrllib._string_list___getitem__(self, *args)

    def __setitem__(self, *args):
        return _pyqrllib._string_list___setitem__(self, *args)

    def pop(self):
        return _pyqrllib._string_list_pop(self)

    def append(self, x):
        return _pyqrllib._string_list_append(self, x)

    def empty(self):
        return _pyqrllib._string_list_empty(self)

    def size(self):
        return _pyqrllib._string_list_size(self)

    def swap(self, v):
        return _pyqrllib._string_list_swap(self, v)

    def begin(self):
        return _pyqrllib._string_list_begin(self)

    def end(self):
        return _pyqrllib._string_list_end(self)

    def rbegin(self):
        return _pyqrllib._string_list_rbegin(self)

    def rend(self):
        return _pyqrllib._string_list_rend(self)

    def clear(self):
        return _pyqrllib._string_list_clear(self)

    def get_allocator(self):
        return _pyqrllib._string_list_get_allocator(self)

    def pop_back(self):
        return _pyqrllib._string_list_pop_back(self)

    def erase(self, *args):
        return _pyqrllib._string_list_erase(self, *args)

    def __init__(self, *args):
        this = _pyqrllib.new__string_list(*args)
        try:
            self.this.append(this)
        except Exception:
            self.this = this

    def push_back(self, x):
        return _pyqrllib._string_list_push_back(self, x)

    def front(self):
        return _pyqrllib._string_list_front(self)

    def back(self):
        return _pyqrllib._string_list_back(self)

    def assign(self, n, x):
        return _pyqrllib._string_list_assign(self, n, x)

    def resize(self, *args):
        return _pyqrllib._string_list_resize(self, *args)

    def insert(self, *args):
        return _pyqrllib._string_list_insert(self, *args)

    def reserve(self, n):
        return _pyqrllib._string_list_reserve(self, n)

    def capacity(self):
        return _pyqrllib._string_list_capacity(self)
    __swig_destroy__ = _pyqrllib.delete__string_list
    __del__ = lambda self: None
_string_list_swigregister = _pyqrllib._string_list_swigregister
_string_list_swigregister(_string_list)

class _string_list_list(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, _string_list_list, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, _string_list_list, name)
    __repr__ = _swig_repr

    def iterator(self):
        return _pyqrllib._string_list_list_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self):
        return _pyqrllib._string_list_list___nonzero__(self)

    def __bool__(self):
        return _pyqrllib._string_list_list___bool__(self)

    def __len__(self):
        return _pyqrllib._string_list_list___len__(self)

    def __getslice__(self, i, j):
        return _pyqrllib._string_list_list___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _pyqrllib._string_list_list___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _pyqrllib._string_list_list___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _pyqrllib._string_list_list___delitem__(self, *args)

    def __getitem__(self, *args):
        return _pyqrllib._string_list_list___getitem__(self, *args)

    def __setitem__(self, *args):
        return _pyqrllib._string_list_list___setitem__(self, *args)

    def pop(self):
        return _pyqrllib._string_list_list_pop(self)

    def append(self, x):
        return _pyqrllib._string_list_list_append(self, x)

    def empty(self):
        return _pyqrllib._string_list_list_empty(self)

    def size(self):
        return _pyqrllib._string_list_list_size(self)

    def swap(self, v):
        return _pyqrllib._string_list_list_swap(self, v)

    def begin(self):
        return _pyqrllib._string_list_list_begin(self)

    def end(self):
        return _pyqrllib._string_list_list_end(self)

    def rbegin(self):
        return _pyqrllib._string_list_list_rbegin(self)

    def rend(self):
        return _pyqrllib._string_list_list_rend(self)

    def clear(self):
        return _pyqrllib._string_list_list_clear(self)

    def get_allocator(self):
        return _pyqrllib._string_list_list_get_allocator(self)

    def pop_back(self):
        return _pyqrllib._string_list_list_pop_back(self)

    def erase(self, *args):
        return _pyqrllib._string_list_list_erase(self, *args)

    def __init__(self, *args):
        this = _pyqrllib.new__string_list_list(*args)
        try:
            self.this.append(this)
        except Exception:
            self.this = this

    def push_back(self, x):
        return _pyqrllib._string_list_list_push_back(self, x)

    def front(self):
        return _pyqrllib._string_list_list_front(self)

    def back(self):
        return _pyqrllib._string_list_list_back(self)

    def assign(self, n, x):
        return _pyqrllib._string_list_list_assign(self, n, x)

    def resize(self, *args):
        return _pyqrllib._string_list_list_resize(self, *args)

    def insert(self, *args):
        return _pyqrllib._string_list_list_insert(self, *args)

    def reserve(self, n):
        return _pyqrllib._string_list_list_reserve(self, n)

    def capacity(self):
        return _pyqrllib._string_list_list_capacity(self)
    __swig_destroy__ = _pyqrllib.delete__string_list_list
    __del__ = lambda self: None
_string_list_list_swigregister = _pyqrllib._string_list_list_swigregister
_string_list_list_swigregister(_string_list_list)


_pyqrllib.ADDRESS_HASH_SIZE_swigconstant(_pyqrllib)
ADDRESS_HASH_SIZE = _pyqrllib.ADDRESS_HASH_SIZE

def bin2hstr(*args):
    return _pyqrllib.bin2hstr(*args)
bin2hstr = _pyqrllib.bin2hstr

def str2bin(s):
    return _pyqrllib.str2bin(s)
str2bin = _pyqrllib.str2bin

def hstr2bin(s):
    return _pyqrllib.hstr2bin(s)
hstr2bin = _pyqrllib.hstr2bin

def bin2mnemonic(vec):
    return _pyqrllib.bin2mnemonic(vec)
bin2mnemonic = _pyqrllib.bin2mnemonic

def mnemonic2bin(mnemonic):
    return _pyqrllib.mnemonic2bin(mnemonic)
mnemonic2bin = _pyqrllib.mnemonic2bin

def getRandomSeed(seed_size, entropy):
    return _pyqrllib.getRandomSeed(seed_size, entropy)
getRandomSeed = _pyqrllib.getRandomSeed

def getHashChainSeed(seed, seed_shift, count):
    return _pyqrllib.getHashChainSeed(seed, seed_shift, count)
getHashChainSeed = _pyqrllib.getHashChainSeed

def sha2_256(input):
    return _pyqrllib.sha2_256(input)
sha2_256 = _pyqrllib.sha2_256

def sha2_256_n(input, count):
    return _pyqrllib.sha2_256_n(input, count)
sha2_256_n = _pyqrllib.sha2_256_n

def shake128(hash_size, input):
    return _pyqrllib.shake128(hash_size, input)
shake128 = _pyqrllib.shake128

def shake256(hash_size, data):
    return _pyqrllib.shake256(hash_size, data)
shake256 = _pyqrllib.shake256

_pyqrllib.SHA2_256_swigconstant(_pyqrllib)
SHA2_256 = _pyqrllib.SHA2_256

_pyqrllib.SHAKE_128_swigconstant(_pyqrllib)
SHAKE_128 = _pyqrllib.SHAKE_128

_pyqrllib.SHAKE_256_swigconstant(_pyqrllib)
SHAKE_256 = _pyqrllib.SHAKE_256

_pyqrllib.SHA256_2X_swigconstant(_pyqrllib)
SHA256_2X = _pyqrllib.SHA256_2X

_pyqrllib.XMSS_swigconstant(_pyqrllib)
XMSS = _pyqrllib.XMSS
class QRLDescriptor(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, QRLDescriptor, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, QRLDescriptor, name)
    __repr__ = _swig_repr

    def __init__(self, hashFunction, signatureType, height, addrFormatType):
        this = _pyqrllib.new_QRLDescriptor(hashFunction, signatureType, height, addrFormatType)
        try:
            self.this.append(this)
        except Exception:
            self.this = this

    def getHashFunction(self):
        return _pyqrllib.QRLDescriptor_getHashFunction(self)

    def getSignatureType(self):
        return _pyqrllib.QRLDescriptor_getSignatureType(self)

    def getHeight(self):
        return _pyqrllib.QRLDescriptor_getHeight(self)

    def getAddrFormatType(self):
        return _pyqrllib.QRLDescriptor_getAddrFormatType(self)
    __swig_getmethods__["fromExtendedSeed"] = lambda x: _pyqrllib.QRLDescriptor_fromExtendedSeed
    if _newclass:
        fromExtendedSeed = staticmethod(_pyqrllib.QRLDescriptor_fromExtendedSeed)
    __swig_getmethods__["fromExtendedPK"] = lambda x: _pyqrllib.QRLDescriptor_fromExtendedPK
    if _newclass:
        fromExtendedPK = staticmethod(_pyqrllib.QRLDescriptor_fromExtendedPK)
    __swig_getmethods__["fromBytes"] = lambda x: _pyqrllib.QRLDescriptor_fromBytes
    if _newclass:
        fromBytes = staticmethod(_pyqrllib.QRLDescriptor_fromBytes)
    __swig_getmethods__["getSize"] = lambda x: _pyqrllib.QRLDescriptor_getSize
    if _newclass:
        getSize = staticmethod(_pyqrllib.QRLDescriptor_getSize)

    def getBytes(self):
        return _pyqrllib.QRLDescriptor_getBytes(self)
    __swig_destroy__ = _pyqrllib.delete_QRLDescriptor
    __del__ = lambda self: None
QRLDescriptor_swigregister = _pyqrllib.QRLDescriptor_swigregister
QRLDescriptor_swigregister(QRLDescriptor)

def QRLDescriptor_fromExtendedSeed(extended_seed):
    return _pyqrllib.QRLDescriptor_fromExtendedSeed(extended_seed)
QRLDescriptor_fromExtendedSeed = _pyqrllib.QRLDescriptor_fromExtendedSeed

def QRLDescriptor_fromExtendedPK(extended_pk):
    return _pyqrllib.QRLDescriptor_fromExtendedPK(extended_pk)
QRLDescriptor_fromExtendedPK = _pyqrllib.QRLDescriptor_fromExtendedPK

def QRLDescriptor_fromBytes(bytes):
    return _pyqrllib.QRLDescriptor_fromBytes(bytes)
QRLDescriptor_fromBytes = _pyqrllib.QRLDescriptor_fromBytes

def QRLDescriptor_getSize():
    return _pyqrllib.QRLDescriptor_getSize()
QRLDescriptor_getSize = _pyqrllib.QRLDescriptor_getSize


_pyqrllib.PICOSHA2_BUFFER_SIZE_FOR_INPUT_ITERATOR_swigconstant(_pyqrllib)
PICOSHA2_BUFFER_SIZE_FOR_INPUT_ITERATOR = _pyqrllib.PICOSHA2_BUFFER_SIZE_FOR_INPUT_ITERATOR

def mask_8bit(x):
    return _pyqrllib.mask_8bit(x)
mask_8bit = _pyqrllib.mask_8bit

def mask_32bit(x):
    return _pyqrllib.mask_32bit(x)
mask_32bit = _pyqrllib.mask_32bit

def ch(x, y, z):
    return _pyqrllib.ch(x, y, z)
ch = _pyqrllib.ch

def maj(x, y, z):
    return _pyqrllib.maj(x, y, z)
maj = _pyqrllib.maj

def rotr(x, n):
    return _pyqrllib.rotr(x, n)
rotr = _pyqrllib.rotr

def bsig0(x):
    return _pyqrllib.bsig0(x)
bsig0 = _pyqrllib.bsig0

def bsig1(x):
    return _pyqrllib.bsig1(x)
bsig1 = _pyqrllib.bsig1

def shr(x, n):
    return _pyqrllib.shr(x, n)
shr = _pyqrllib.shr

def ssig0(x):
    return _pyqrllib.ssig0(x)
ssig0 = _pyqrllib.ssig0

def ssig1(x):
    return _pyqrllib.ssig1(x)
ssig1 = _pyqrllib.ssig1
class hash256_one_by_one(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, hash256_one_by_one, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, hash256_one_by_one, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _pyqrllib.new_hash256_one_by_one()
        try:
            self.this.append(this)
        except Exception:
            self.this = this

    def init(self):
        return _pyqrllib.hash256_one_by_one_init(self)

    def finish(self):
        return _pyqrllib.hash256_one_by_one_finish(self)
    __swig_destroy__ = _pyqrllib.delete_hash256_one_by_one
    __del__ = lambda self: None
hash256_one_by_one_swigregister = _pyqrllib.hash256_one_by_one_swigregister
hash256_one_by_one_swigregister(hash256_one_by_one)
cvar = _pyqrllib.cvar
add_constant = cvar.add_constant
initial_message_digest = cvar.initial_message_digest


def get_hash_hex_string(*args):
    return _pyqrllib.get_hash_hex_string(*args)
get_hash_hex_string = _pyqrllib.get_hash_hex_string

def hash256_hex_string(src, hex_str):
    return _pyqrllib.hash256_hex_string(src, hex_str)
hash256_hex_string = _pyqrllib.hash256_hex_string
class QRLHelper(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, QRLHelper, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, QRLHelper, name)
    __repr__ = _swig_repr

    def __init__(self):
        this = _pyqrllib.new_QRLHelper()
        try:
            self.this.append(this)
        except Exception:
            self.this = this
    __swig_getmethods__["getAddress"] = lambda x: _pyqrllib.QRLHelper_getAddress
    if _newclass:
        getAddress = staticmethod(_pyqrllib.QRLHelper_getAddress)
    __swig_getmethods__["addressIsValid"] = lambda x: _pyqrllib.QRLHelper_addressIsValid
    if _newclass:
        addressIsValid = staticmethod(_pyqrllib.QRLHelper_addressIsValid)
    __swig_destroy__ = _pyqrllib.delete_QRLHelper
    __del__ = lambda self: None
QRLHelper_swigregister = _pyqrllib.QRLHelper_swigregister
QRLHelper_swigregister(QRLHelper)

def QRLHelper_getAddress(extended_pk):
    return _pyqrllib.QRLHelper_getAddress(extended_pk)
QRLHelper_getAddress = _pyqrllib.QRLHelper_getAddress

def QRLHelper_addressIsValid(address):
    return _pyqrllib.QRLHelper_addressIsValid(address)
QRLHelper_addressIsValid = _pyqrllib.QRLHelper_addressIsValid

class xmss_params(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, xmss_params, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, xmss_params, name)
    __repr__ = _swig_repr
    __swig_setmethods__["wots_par"] = _pyqrllib.xmss_params_wots_par_set
    __swig_getmethods__["wots_par"] = _pyqrllib.xmss_params_wots_par_get
    if _newclass:
        wots_par = _swig_property(_pyqrllib.xmss_params_wots_par_get, _pyqrllib.xmss_params_wots_par_set)
    __swig_setmethods__["n"] = _pyqrllib.xmss_params_n_set
    __swig_getmethods__["n"] = _pyqrllib.xmss_params_n_get
    if _newclass:
        n = _swig_property(_pyqrllib.xmss_params_n_get, _pyqrllib.xmss_params_n_set)
    __swig_setmethods__["h"] = _pyqrllib.xmss_params_h_set
    __swig_getmethods__["h"] = _pyqrllib.xmss_params_h_get
    if _newclass:
        h = _swig_property(_pyqrllib.xmss_params_h_get, _pyqrllib.xmss_params_h_set)
    __swig_setmethods__["k"] = _pyqrllib.xmss_params_k_set
    __swig_getmethods__["k"] = _pyqrllib.xmss_params_k_get
    if _newclass:
        k = _swig_property(_pyqrllib.xmss_params_k_get, _pyqrllib.xmss_params_k_set)

    def __init__(self):
        this = _pyqrllib.new_xmss_params()
        try:
            self.this.append(this)
        except Exception:
            self.this = this
    __swig_destroy__ = _pyqrllib.delete_xmss_params
    __del__ = lambda self: None
xmss_params_swigregister = _pyqrllib.xmss_params_swigregister
xmss_params_swigregister(xmss_params)

class XmssBase(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, XmssBase, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, XmssBase, name)

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined - class is abstract")
    __repr__ = _swig_repr
    __swig_destroy__ = _pyqrllib.delete_XmssBase
    __del__ = lambda self: None

    def sign(self, message):
        return _pyqrllib.XmssBase_sign(self, message)
    __swig_getmethods__["verify"] = lambda x: _pyqrllib.XmssBase_verify
    if _newclass:
        verify = staticmethod(_pyqrllib.XmssBase_verify)

    def getSK(self):
        return _pyqrllib.XmssBase_getSK(self)

    def getPK(self):
        return _pyqrllib.XmssBase_getPK(self)

    def getDescriptor(self):
        return _pyqrllib.XmssBase_getDescriptor(self)

    def getDescriptorBytes(self):
        return _pyqrllib.XmssBase_getDescriptorBytes(self)

    def getHeight(self):
        return _pyqrllib.XmssBase_getHeight(self)

    def getSeed(self):
        return _pyqrllib.XmssBase_getSeed(self)

    def getExtendedSeed(self):
        return _pyqrllib.XmssBase_getExtendedSeed(self)

    def getRoot(self):
        return _pyqrllib.XmssBase_getRoot(self)

    def getPKSeed(self):
        return _pyqrllib.XmssBase_getPKSeed(self)

    def getSKSeed(self):
        return _pyqrllib.XmssBase_getSKSeed(self)

    def getSKPRF(self):
        return _pyqrllib.XmssBase_getSKPRF(self)

    def getAddress(self):
        return _pyqrllib.XmssBase_getAddress(self)

    def getNumberSignatures(self):
        return _pyqrllib.XmssBase_getNumberSignatures(self)

    def getRemainingSignatures(self):
        return _pyqrllib.XmssBase_getRemainingSignatures(self)

    def getIndex(self):
        return _pyqrllib.XmssBase_getIndex(self)

    def setIndex(self, new_index):
        return _pyqrllib.XmssBase_setIndex(self, new_index)

    def getSignatureSize(self):
        return _pyqrllib.XmssBase_getSignatureSize(self)
    __swig_getmethods__["getHeightFromSigSize"] = lambda x: _pyqrllib.XmssBase_getHeightFromSigSize
    if _newclass:
        getHeightFromSigSize = staticmethod(_pyqrllib.XmssBase_getHeightFromSigSize)

    def getSecretKeySize(self):
        return _pyqrllib.XmssBase_getSecretKeySize(self)

    def getPublicKeySize(self):
        return _pyqrllib.XmssBase_getPublicKeySize(self)
XmssBase_swigregister = _pyqrllib.XmssBase_swigregister
XmssBase_swigregister(XmssBase)

def XmssBase_verify(message, signature, pk):
    return _pyqrllib.XmssBase_verify(message, signature, pk)
XmssBase_verify = _pyqrllib.XmssBase_verify

def XmssBase_getHeightFromSigSize(sigSize):
    return _pyqrllib.XmssBase_getHeightFromSigSize(sigSize)
XmssBase_getHeightFromSigSize = _pyqrllib.XmssBase_getHeightFromSigSize

class XmssBasic(XmssBase):
    __swig_setmethods__ = {}
    for _s in [XmssBase]:
        __swig_setmethods__.update(getattr(_s, '__swig_setmethods__', {}))
    __setattr__ = lambda self, name, value: _swig_setattr(self, XmssBasic, name, value)
    __swig_getmethods__ = {}
    for _s in [XmssBase]:
        __swig_getmethods__.update(getattr(_s, '__swig_getmethods__', {}))
    __getattr__ = lambda self, name: _swig_getattr(self, XmssBasic, name)
    __repr__ = _swig_repr

    def __init__(self, seed, height, hashFunction, addrFormatType):
        this = _pyqrllib.new_XmssBasic(seed, height, hashFunction, addrFormatType)
        try:
            self.this.append(this)
        except Exception:
            self.this = this

    def sign(self, message):
        return _pyqrllib.XmssBasic_sign(self, message)
    __swig_destroy__ = _pyqrllib.delete_XmssBasic
    __del__ = lambda self: None
XmssBasic_swigregister = _pyqrllib.XmssBasic_swigregister
XmssBasic_swigregister(XmssBasic)

class treehash_inst(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, treehash_inst, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, treehash_inst, name)
    __repr__ = _swig_repr
    __swig_setmethods__["h"] = _pyqrllib.treehash_inst_h_set
    __swig_getmethods__["h"] = _pyqrllib.treehash_inst_h_get
    if _newclass:
        h = _swig_property(_pyqrllib.treehash_inst_h_get, _pyqrllib.treehash_inst_h_set)
    __swig_setmethods__["next_idx"] = _pyqrllib.treehash_inst_next_idx_set
    __swig_getmethods__["next_idx"] = _pyqrllib.treehash_inst_next_idx_get
    if _newclass:
        next_idx = _swig_property(_pyqrllib.treehash_inst_next_idx_get, _pyqrllib.treehash_inst_next_idx_set)
    __swig_setmethods__["stackusage"] = _pyqrllib.treehash_inst_stackusage_set
    __swig_getmethods__["stackusage"] = _pyqrllib.treehash_inst_stackusage_get
    if _newclass:
        stackusage = _swig_property(_pyqrllib.treehash_inst_stackusage_get, _pyqrllib.treehash_inst_stackusage_set)
    __swig_setmethods__["completed"] = _pyqrllib.treehash_inst_completed_set
    __swig_getmethods__["completed"] = _pyqrllib.treehash_inst_completed_get
    if _newclass:
        completed = _swig_property(_pyqrllib.treehash_inst_completed_get, _pyqrllib.treehash_inst_completed_set)
    __swig_setmethods__["node"] = _pyqrllib.treehash_inst_node_set
    __swig_getmethods__["node"] = _pyqrllib.treehash_inst_node_get
    if _newclass:
        node = _swig_property(_pyqrllib.treehash_inst_node_get, _pyqrllib.treehash_inst_node_set)

    def __init__(self):
        this = _pyqrllib.new_treehash_inst()
        try:
            self.this.append(this)
        except Exception:
            self.this = this
    __swig_destroy__ = _pyqrllib.delete_treehash_inst
    __del__ = lambda self: None
treehash_inst_swigregister = _pyqrllib.treehash_inst_swigregister
treehash_inst_swigregister(treehash_inst)

class bds_state(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, bds_state, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, bds_state, name)
    __repr__ = _swig_repr
    __swig_setmethods__["stack"] = _pyqrllib.bds_state_stack_set
    __swig_getmethods__["stack"] = _pyqrllib.bds_state_stack_get
    if _newclass:
        stack = _swig_property(_pyqrllib.bds_state_stack_get, _pyqrllib.bds_state_stack_set)
    __swig_setmethods__["stackoffset"] = _pyqrllib.bds_state_stackoffset_set
    __swig_getmethods__["stackoffset"] = _pyqrllib.bds_state_stackoffset_get
    if _newclass:
        stackoffset = _swig_property(_pyqrllib.bds_state_stackoffset_get, _pyqrllib.bds_state_stackoffset_set)
    __swig_setmethods__["stacklevels"] = _pyqrllib.bds_state_stacklevels_set
    __swig_getmethods__["stacklevels"] = _pyqrllib.bds_state_stacklevels_get
    if _newclass:
        stacklevels = _swig_property(_pyqrllib.bds_state_stacklevels_get, _pyqrllib.bds_state_stacklevels_set)
    __swig_setmethods__["auth"] = _pyqrllib.bds_state_auth_set
    __swig_getmethods__["auth"] = _pyqrllib.bds_state_auth_get
    if _newclass:
        auth = _swig_property(_pyqrllib.bds_state_auth_get, _pyqrllib.bds_state_auth_set)
    __swig_setmethods__["keep"] = _pyqrllib.bds_state_keep_set
    __swig_getmethods__["keep"] = _pyqrllib.bds_state_keep_get
    if _newclass:
        keep = _swig_property(_pyqrllib.bds_state_keep_get, _pyqrllib.bds_state_keep_set)
    __swig_setmethods__["treehash"] = _pyqrllib.bds_state_treehash_set
    __swig_getmethods__["treehash"] = _pyqrllib.bds_state_treehash_get
    if _newclass:
        treehash = _swig_property(_pyqrllib.bds_state_treehash_get, _pyqrllib.bds_state_treehash_set)
    __swig_setmethods__["retain"] = _pyqrllib.bds_state_retain_set
    __swig_getmethods__["retain"] = _pyqrllib.bds_state_retain_get
    if _newclass:
        retain = _swig_property(_pyqrllib.bds_state_retain_get, _pyqrllib.bds_state_retain_set)
    __swig_setmethods__["next_leaf"] = _pyqrllib.bds_state_next_leaf_set
    __swig_getmethods__["next_leaf"] = _pyqrllib.bds_state_next_leaf_get
    if _newclass:
        next_leaf = _swig_property(_pyqrllib.bds_state_next_leaf_get, _pyqrllib.bds_state_next_leaf_set)

    def __init__(self):
        this = _pyqrllib.new_bds_state()
        try:
            self.this.append(this)
        except Exception:
            self.this = this
    __swig_destroy__ = _pyqrllib.delete_bds_state
    __del__ = lambda self: None
bds_state_swigregister = _pyqrllib.bds_state_swigregister
bds_state_swigregister(bds_state)


def xmss_set_bds_state(state, stack, stackoffset, stacklevels, auth, keep, treehash, retain, next_leaf):
    return _pyqrllib.xmss_set_bds_state(state, stack, stackoffset, stacklevels, auth, keep, treehash, retain, next_leaf)
xmss_set_bds_state = _pyqrllib.xmss_set_bds_state

def xmssfast_Genkeypair(hash_func, params, pk, sk, state, seed):
    return _pyqrllib.xmssfast_Genkeypair(hash_func, params, pk, sk, state, seed)
xmssfast_Genkeypair = _pyqrllib.xmssfast_Genkeypair

def xmssfast_Signmsg(hash_func, params, sk, state, sig_msg, msg, msglen):
    return _pyqrllib.xmssfast_Signmsg(hash_func, params, sk, state, sig_msg, msg, msglen)
xmssfast_Signmsg = _pyqrllib.xmssfast_Signmsg

def xmssfast_update(hash_func, params, sk, state, new_idx):
    return _pyqrllib.xmssfast_update(hash_func, params, sk, state, new_idx)
xmssfast_update = _pyqrllib.xmssfast_update
class XmssFast(XmssBase):
    __swig_setmethods__ = {}
    for _s in [XmssBase]:
        __swig_setmethods__.update(getattr(_s, '__swig_setmethods__', {}))
    __setattr__ = lambda self, name, value: _swig_setattr(self, XmssFast, name, value)
    __swig_getmethods__ = {}
    for _s in [XmssBase]:
        __swig_getmethods__.update(getattr(_s, '__swig_getmethods__', {}))
    __getattr__ = lambda self, name: _swig_getattr(self, XmssFast, name)
    __repr__ = _swig_repr

    def __init__(self, *args):
        this = _pyqrllib.new_XmssFast(*args)
        try:
            self.this.append(this)
        except Exception:
            self.this = this

    def sign(self, message):
        return _pyqrllib.XmssFast_sign(self, message)

    def setIndex(self, new_index):
        return _pyqrllib.XmssFast_setIndex(self, new_index)
    __swig_destroy__ = _pyqrllib.delete_XmssFast
    __del__ = lambda self: None
XmssFast_swigregister = _pyqrllib.XmssFast_swigregister
XmssFast_swigregister(XmssFast)

class XmssPool(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, XmssPool, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, XmssPool, name)
    __repr__ = _swig_repr

    def __init__(self, base_seed, height, starting_index, pool_size):
        this = _pyqrllib.new_XmssPool(base_seed, height, starting_index, pool_size)
        try:
            self.this.append(this)
        except Exception:
            self.this = this
    __swig_destroy__ = _pyqrllib.delete_XmssPool
    __del__ = lambda self: None

    def getNextTree(self):
        return _pyqrllib.XmssPool_getNextTree(self)

    def isAvailable(self):
        return _pyqrllib.XmssPool_isAvailable(self)

    def getCurrentIndex(self):
        return _pyqrllib.XmssPool_getCurrentIndex(self)
XmssPool_swigregister = _pyqrllib.XmssPool_swigregister
XmssPool_swigregister(XmssPool)

# This file is compatible with both classic and new-style classes.





# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from __future__ import print_function
from unittest import TestCase

from pyqrllib import pyqrllib


class TestShake128(TestCase):
    # Values obtained from https://emn178.github.io/online-tools/shake_128.html
    shake128_input1 = 'hello'
    shake128_expected_result1 = '8eb4b6a932f280335ee1a279f8c208a349e7bc65daf831d3021c213825292463'
    shake128_input2 = 'hello-qrl'
    shake128_expected_result2 = '50028af4e91b430a1ec24924edc707b0d24ab01be44ea5f5c5c111087e9aadcb'

    def __init__(self, *args, **kwargs):
        super(TestShake128, self).__init__(*args, **kwargs)

    def check_shake_result(self, data_text, expected):
        hex_in_before = pyqrllib.bin2hstr(pyqrllib.str2bin(data_text))
        data_out = pyqrllib.shake128(32, pyqrllib.str2bin(data_text))

        # This is just to keep as an example. Things could be compared without converting to hex
        hex_in = pyqrllib.bin2hstr(pyqrllib.str2bin(data_text))
        hex_out = pyqrllib.bin2hstr(data_out)

        self.assertEqual(hex_in, hex_in_before)
        self.assertEqual(hex_out, expected)

    def test_check_shake128(self):
        self.check_shake_result(self.shake128_input1, self.shake128_expected_result1)
        self.check_shake_result(self.shake128_input2, self.shake128_expected_result2)


class TestShake256(TestCase):
    # values obtained from https://emn178.github.io/online-tools/shake_256.html
    shake256_input1 = 'hello'
    shake256_expected_result1 = '1234075ae4a1e77316cf2d8000974581a343b9ebbc' \
                                'a7e3d1db83394c30f221626f594e4f0de63902349a' \
                                '5ea5781213215813919f92a4d86d127466e3d07e8be3'

    shake256_input2 = 'hello-1234'
    shake256_expected_result2 = '4a01ca14fd8468f2d2e3a0b3d7597731ad15501675' \
                                '3677807ed735b022a9944e61586a6378fc6ffe49e9' \
                                'e0e456f8e2bbfaa41330c5ae7005a2d24ac8f0597e60'

    def __init__(self, *args, **kwargs):
        super(TestShake256, self).__init__(*args, **kwargs)

    def check_shake_result(self, data_text, expected):
        # Move data into array
        hex_in_before = pyqrllib.bin2hstr(pyqrllib.str2bin(data_text))
        data_out = pyqrllib.shake256(64, pyqrllib.str2bin(data_text))

        # This is just to keep as an example. Things could be compared without converting to hex
        hex_in = pyqrllib.bin2hstr(pyqrllib.str2bin(data_text))
        hex_out = pyqrllib.bin2hstr(data_out)

        self.assertEqual(hex_in, hex_in_before)
        self.assertEqual(hex_out, expected)

    def test_check_shake256(self):
        self.check_shake_result(self.shake256_input1,
                                self.shake256_expected_result1)
        self.check_shake_result(self.shake256_input2,
                                self.shake256_expected_result2)



from math import log
from Kyber_failure import p2_cyclotomic_error_probability
from MLWE_security import MLWE_summarize_attacks, MLWEParameterSet


class KyberParameterSet:
    def __init__(self, n, m, ks, ke,  q, rqk, rqc, rq2):
        self.n = n
        self.m = m
        self.ks= ks     # binary distribution for the secret key
        self.ke = ke    # binary distribution for the ciphertext errors
        self.q = q 
        self.rqk = rqk  # 2^(bits in the public key)  
        self.rqc = rqc  # 2^(bits in the first ciphertext)
        self.rq2 = rq2  # 2^(bits in the second ciphertext)


def Kyber_to_MLWE(kps):
    if kps.ks != kps.ke:
        raise "The security script does not handle different error parameter in secrets and errors (ks != ke) "

    return MLWEParameterSet(kps.n, kps.m, kps.m + 1, kps.ks, kps.q)


def communication_costs(ps):
    """ Compute the communication cost of a parameter set
    :param ps: Parameter set (ParameterSet)
    :returns: (cost_Alice, cost_Bob) (in Bytes)
    """    
    A_space = 256 + ps.n * ps.m * log(ps.rqk)/log(2)
    B_space = ps.n * ps.m * log(ps.rqc)/log(2) + ps.n * log(ps.rq2)/log(2) + 256
    return (int(round(A_space))/8., int(round(B_space))/8.)


def summarize(ps):
    print ("params: ", ps.__dict__)
    print ("com costs: ", communication_costs(ps))
    F, f = p2_cyclotomic_error_probability(ps)
    print ("failure: %.1f = 2^%.1f"%(f, log(f + 2.**(-300))/log(2)))


if __name__ == "__main__":
    # Parameter sets
    ps_recommended = KyberParameterSet(256, 3, 4, 4, 7681, 2**11, 2**11, 2**3)
    ps_light = KyberParameterSet(256, 2, 5, 5, 7681, 2**11, 2**11, 2**3)
    ps_paranoid = KyberParameterSet(256, 4, 3, 3, 7681, 2**11, 2**11, 2**3)

    # Analyses
    print ("Kyber512 (light):")
    print ("--------------------")
    print ("security:")
    MLWE_summarize_attacks(Kyber_to_MLWE(ps_light))
    summarize(ps_light)
    print ()

    print ("Kyber768 (recommended):")
    print ("--------------------")
    print ("security:")
    MLWE_summarize_attacks(Kyber_to_MLWE(ps_recommended))
    summarize(ps_recommended)
    print ()

    print ("Kyber1024 (paranoid):")
    print ("--------------------")
    print ("security:")
    MLWE_summarize_attacks(Kyber_to_MLWE(ps_paranoid))
    summarize(ps_paranoid)
    print ()
